"""
The main Stretchy Tree Matching algorithm, implemented as a class.
"""
import ast
import re
from pedal.cait.ast_map import AstMap
from pedal.cait.cait_node import CaitNode
from pedal.cait.ast_map import SymTables

# "Enums" for _name_regex
_VAR = "var"
_EXP = "exp"
_WILD = "wild"
_NONE_FIELD = "none"


def is_primitive(item):
    """
    Determines if the given item is a primitive value (either an int, float,
    str, bool, or None).

    Args:
        item (any): Any value
    Returns:
        bool: Whether the item is a primitive value.
    """
    return isinstance(item, (int, float, str, bool)) or item is None


def _name_regex(name_id):
    var_match = re.compile('^_[^_].*_$')  # /regex
    exp_match = re.compile('^__.*__$')  # /regex
    wild_card = re.compile('^___$')  # /regex
    return {_VAR: var_match.match(name_id),
            _EXP: exp_match.match(name_id),
            _WILD: wild_card.match(name_id)}


class StretchyTreeMatcher:
    """
    The StretchyTreeMatcher is used to compare a pattern against some
    student code. It produces a set of potential mappings between them.

    Args:
        ast_or_code (str or AstNode): The students' code or a valid AstNode from
            `ast.parse`. If the code has invalid syntax, a SyntaxError
            will be raised.
        filename (str): The filename to parse with - only used for error
            reporting.
        report (Report): A report to obtain data from.
    """
    def __init__(self, ast_or_code, report, filename="__main__"):
        self.report = report
        if isinstance(ast_or_code, str):
            ast_node = ast.parse(ast_or_code, filename)
        else:
            ast_node = ast_or_code
        # Build up root
        if ast_node is None:
            self.root_node = None
        elif isinstance(ast_node, CaitNode):
            self.root_node = ast_node
        else:
            self.root_node = CaitNode(ast_node, _NONE_FIELD, report=self.report)

    def find_matches(self, ast_or_code, filename="__main__", check_meta=True, use_previous=None):
        """
        Args:
            use_previous:
            ast_or_code (str or AstNode): The students' code or a valid AstNode from
                `ast.parse`. If the code has invalid syntax, a SyntaxError
                will be raised.
            filename (str): The filename to parse with - only used for error
                reporting.
            check_meta (bool): Determine if the nodes came from the same AST
                field.
        Returns:
            list[AstMap]: A list of AstMaps that are suitable matches.
        """
        if isinstance(ast_or_code, str):
            other_tree = CaitNode(ast.parse(ast_or_code, filename), report=self.report)
        elif isinstance(ast_or_code, CaitNode):
            other_tree = ast_or_code
        else:
            other_tree = CaitNode(ast_or_code, _NONE_FIELD, report=self.report)
        explore_root = self.root_node
        trim_set = ["Expr", "Module"]
        explore_root_old_field = explore_root.field
        if self.root_node is not None:  # Trimming ins_node
            while (len(explore_root.children) == 1 and
                   explore_root.ast_name in trim_set):
                explore_root.field = explore_root_old_field
                explore_root = explore_root.children[0]
                explore_root_old_field = explore_root.field
                explore_root.field = _NONE_FIELD
        other_root = other_tree
        other_root_old_field = other_root.field
        if other_root is not None:  # Trimming std_node
            while len(other_root.children) == 1 and other_root.ast_name in trim_set:
                other_root.field = other_root_old_field
                other_root = other_root.children[0]
                other_root_old_field = other_root.field
                other_root.field = _NONE_FIELD
        matches = self.any_node_match(explore_root, other_root,
                                      check_meta=check_meta, use_previous=use_previous)
        explore_root.field = explore_root_old_field
        other_root.field = other_root_old_field
        return matches

    def any_node_match(self, ins_node, std_node, check_meta=True, cut=False, use_previous=None):
        """
        Finds whether ins_node can be matched to some node in the tree std_node

        Args:
            use_previous:
            ins_node:
            std_node:
            check_meta:
            cut:

        Returns:
            list of AstMaps: a mapping of nodes and a symbol table mapping ins_node to
                some node in the tree std_node or False if such a matching does not
                exist
        """
        # @TODO: create a more public function that converts ins_node and std_node into CaitNodes
        # TODO: Create exhaustive any_node_match
        # matching: an object representing the mapping and the symbol table
        matching = self.deep_find_match(ins_node, std_node, check_meta, use_previous=use_previous)
        # if a direct matching is found
        if matching:
            for match in matching:
                match.match_root = match.mappings[ins_node]
        else:
            matching = []
        #    return matching  # return it
        # if not matching or exhaust:  # otherwise
        # try to matching ins_node to each child of std_node, recursively
        for std_child in std_node.children:
            matching_c = self.any_node_match(ins_node, std_child, check_meta=check_meta, cut=cut, use_previous=use_previous)
            if matching_c:
                for match in matching_c:
                    match.match_root = match.mappings[ins_node]
                # return matching
                matching = matching + matching_c
        if len(matching) > 0:
            return matching
        return []

    def deep_find_match(self, ins_node, std_node, check_meta=True,
                        use_previous=None):
        """
        Finds whether ins_node and matches std_node and whether ins_node's children flexibly match std_node's children
        in order
        Args:
            ins_node: The instructor ast that should be included in the student AST
            std_node: The student AST that we are searching for the included tree
            check_meta: Flag, if True, check whether the two nodes originated from the same ast field
            use_previous: If this was part of a previous match...

        Returns:
            a mapping of nodes and a symbol table mapping ins_node to std_node, or [] if no mapping was found
        """
        method_name = "deep_find_match_" + type(ins_node.astNode).__name__
        target_func = getattr(self, method_name, self.deep_find_match_generic)
        return target_func(ins_node, std_node, check_meta, use_previous=use_previous)

    # noinspection PyPep8Naming
    def deep_find_match_Name(self, ins_node, std_node, check_meta=True, use_previous=None):
        """

        Args:
            ins_node:
            std_node:
            check_meta:
            use_previous:

        Returns:

        """
        name_id = ins_node.astNode.id
        match = _name_regex(name_id)
        mapping = AstMap()
        matched = False
        meta_matched = self.metas_match(ins_node, std_node, check_meta)
        if match[_VAR] and meta_matched:  # if variable
            if type(std_node.astNode).__name__ == "Name":
                return self.deep_find_match_generic(ins_node, std_node,
                                                    check_meta=check_meta, ignores=["ctx"], use_previous=use_previous)
        # could else return False, but shallow_match_generic should do this as well
        elif match[_EXP] and meta_matched:  # and meta_matched:  # if expression
            # terminate recursion, the whole subtree should match since expression nodes match to anything
            mapping.merge_map_with(use_previous)
            mapping.add_exp_to_sym_table(ins_node, std_node)
            matched = True
        elif match[_WILD] and meta_matched:  # if wild card, don't care
            # terminate the recursion, the whole subtree should match since wild cards match to anything
            matched = True

        if matched:
            mapping.add_node_pairing(ins_node, std_node)
            return [mapping]
        # else
        return self.deep_find_match_generic(ins_node, std_node,
                                            check_meta=check_meta, ignores=["ctx"], use_previous=use_previous)

    # noinspection PyPep8Naming
    def deep_find_match_BinOp(self, ins_node, std_node, check_meta=True, use_previous=None):
        """

        Args:
            ins_node:
            std_node:
            check_meta:
            use_previous:

        Returns:

        """
        op = ins_node.astNode.op
        op = type(op).__name__
        is_generic = not (op == "Mult" or op == "Add")
        if is_generic:
            return self.deep_find_match_generic(ins_node, std_node, check_meta, use_previous=use_previous)
        else:  # this means that the node is clearly commutative
            return self.deep_find_match_binflex(ins_node, std_node, False, use_previous=use_previous)

    # noinspection PyMethodMayBeStatic
    def binflex_helper(self, case_left, case_right, new_mappings, base_mappings, use_previous=None):
        """
        adds to new_mappings (return/modify by argument) the mappings for both the left and right subtrees as denoted by
        case_left and case_right
        Args:
            case_left: The mappings for the left opperand
            case_right: The mappings for the right opperand
            new_mappings: The new set of mappings to generate
            base_mappings: The original mappings of the binop node
            use_previous: A mapping passed down from an initial match
        Returns:
            None
        """
        if case_left and case_right:
            for case_l in case_left:
                new_map = base_mappings[0].new_merged_map(case_l).new_merged_map(use_previous)
                for case_r in case_right:
                    both = new_map.new_merged_map(case_r)
                    if not both.has_conflicts():
                        new_mappings.append(both)

    def deep_find_match_binflex(self, ins_node, std_node, check_meta=False, use_previous=None):
        """

        Args:
            ins_node:
            std_node:
            check_meta:
            use_previous:

        Returns:

        """
        base_mappings = self.shallow_match(ins_node, std_node, check_meta)
        if not base_mappings:
            return []
        op_mappings = self.shallow_match(ins_node.children[1], std_node.children[1], check_meta=True)
        if not op_mappings:
            return []
        base_mappings = [base_mappings[0].new_merged_map(op_mappings[0])]

        if base_mappings:
            ins_left = ins_node.children[0]  # instructor left ast node
            ins_right = ins_node.children[2]  # instructor right ast node
            std_left = std_node.children[0]  # student left ast node
            std_right = std_node.children[2]  # student right ast node
            new_mappings = []
            # case 1: ins_left->std_left and ins_right->std_right
            case_left = self.deep_find_match(ins_left, std_left, False)
            case_right = self.deep_find_match(ins_right, std_right, False)
            self.binflex_helper(case_left, case_right, new_mappings, base_mappings, use_previous=use_previous)
            # case 2: ins_left->std_right and ins_right->std_left
            case_left = self.deep_find_match(ins_left, std_right, False)
            case_right = self.deep_find_match(ins_right, std_left, False)
            self.binflex_helper(case_left, case_right, new_mappings, base_mappings, use_previous=use_previous)
            if len(new_mappings) == 0:
                return []
            return new_mappings
        return []

    def deep_find_match_Expr(self, ins_node, std_node, check_meta=True, use_previous=None):
        """
        An Expression node (not to be confused with expressions denoted by the instructor nodes in Name ast nodes)
        checks whether it should be generic, or not
        Args:
            ins_node: Instructor ast to find in the student ast
            std_node: Student AST to search for the instructor ast in
            check_meta: flag to check whether the fields of the instructor node and the student node should match
            use_previous: An AstMap from a previous matching run

        Returns:
            AstMap: a mapping between the instructor and student asts, or False if such a mapping doesn't exist
        """
        # if check_meta and ins_node.field != std_node.field:
        if not self.metas_match(ins_node, std_node, check_meta):
            return []
        mapping = AstMap() if use_previous is None else use_previous
        value = ins_node.value
        ast_type = type(value.astNode).__name__
        if ast_type == "Name":
            name_id = value.astNode.id
            exp_match = re.compile('^__.*__$')  # /regex
            wild_card = re.compile('^___$')  # /regex
            matched = False
            meta_matched = self.metas_match(ins_node, std_node, check_meta)
            if exp_match.match(name_id) and meta_matched:  # and meta_matched:  # if expression
                # terminate recursion, the whole subtree should match since expression nodes match to anything
                mapping.add_exp_to_sym_table(value, std_node)
                matched = True
            elif wild_card.match(name_id) and meta_matched:  # if wild card, don't care
                # terminate the recursion, the whole subtree should match since wild cards match to anything
                matched = True
            if matched:
                mapping.add_node_pairing(ins_node, std_node)
                return [mapping]
        return self.deep_find_match_generic(ins_node, std_node, check_meta)

    def deep_find_match_generic(self, ins_node, std_node, check_meta=True, ignores=None, use_previous=None):
        """
        This first uses shallow match to find a base map (match) from which to
        build off. The algorithm then tracks all the possible mappings that
        match a given child node in the instructor AST, keeping track of which
        siblings have been visited.

        For each instructor child, when all children of the student node have
        been iterated through recursively, a helper function is called. This
        helper function determines which possible children validly can extend
        the base match to create a set of new base maps through use of the
        indicies of the sibilings.

        The process repeats itself until no matches can be grown or until each
        instructor child node has been visited

        Args:
            ins_node: Instructor ast to find in the student ast
            std_node: Student AST to search for the instructor ast in
            check_meta: flag to check whether the fields of the instructor node and the student node should match
            ignores: List of fields to ignore in the field match
            use_previous: a map from a previous match

        Returns:
            a mapping between the isntructor and student asts, or [] if such a mapping doesn't exist
        """
        if ignores is None:
            ignores = []
        base_mappings = self.shallow_match(ins_node, std_node, check_meta)
        if base_mappings:
            for mapping in base_mappings:
                mapping.merge_map_with(use_previous)
            # base case this runs 0 times because no children
            # find each child of ins_node that matches IN ORDER
            base_sibs = [-1]
            youngest_sib = 0
            # for each child
            for i, insChild in enumerate(ins_node.children):
                # make a new set of maps
                running_maps = []
                running_sibs = []
                if insChild.field in ignores:
                    continue
                # accumulate all potential matches for current child
                for j, std_child in enumerate(std_node.children[youngest_sib:], youngest_sib):
                    std_child = std_node.children[j]
                    new_mapping = self.deep_find_match(insChild, std_child, check_meta)
                    if new_mapping:
                        running_maps.append(new_mapping)
                        running_sibs.append(j)
                map_update = self.map_merge(base_mappings, base_sibs, running_maps, running_sibs)
                if map_update is None:
                    return []
                base_mappings = map_update['new_maps']
                base_sibs = map_update['new_sibs']
                youngest_sib = map_update['youngest_sib'] + 1
            return base_mappings
        return []

    # noinspection PyMethodMayBeStatic
    def map_merge(self, base_maps, base_sibs, run_maps, run_sibs):
        """
        Merges base_maps with the current possible maps. Helper method to deep_find_match_generic. checks whether each
        mapping in run_maps can extend the match to any possible mapping in base_maps.

        Args:
            base_maps: The original mappings
            base_sibs: The corresponding siblings for each mapping in base_maps
            run_maps: The set of maps to merge into the current base_maps
            run_sibs: The corresponding siblings for each mapping in run_maps

        Returns:
            A new set of maps for all valid extensions of base_maps with running maps
        """
        # no matching nodes were found
        if len(run_maps) == 0:
            return None
        new_maps = []
        new_sibs = []
        youngest_sib = run_sibs[0]
        for baseMap, base_sib in zip(base_maps, base_sibs):
            for run_map, runSib in zip(run_maps, run_sibs):
                if runSib > base_sib:
                    for run_mapsub in run_map:
                        new_map = baseMap.new_merged_map(run_mapsub)
                        if not new_map.has_conflicts():  # if it's a valid mapping
                            new_maps.append(new_map)
                            new_sibs.append(runSib)
        if len(new_maps) == 0:
            return None
        return {
            'new_maps': new_maps,
            'new_sibs': new_sibs,
            'youngest_sib': youngest_sib
        }

    # noinspection PyMethodMayBeStatic,PyPep8Naming,PyUnusedLocal
    def shallow_match_Module(self, ins_node, std_node, check_meta=True):
        """
        Flexibly matches a module node to a module or a body
        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:
            a mapping of ins_node to std_node, or False if doesn't match
        """
        if type(std_node.astNode).__name__ == "Module" or std_node.field == "body":
            mapping = AstMap()
            mapping.add_node_pairing(ins_node, std_node)
            return [mapping]
        return []

    def shallow_symbol_handler(self, ins_node, std_node, id_val, check_meta=True):
        """
        TODO: Make this handle the func field to handle functions
        Matches ins_node to std_node for different cases of encountering a name node in ins_node
            case 1: _var_ matches if std_node is a name node and automatically returns a mapping and symbol table
            case 2: __exp__ matches to any subtree and automatically returns a mapping and symbol table
            case 3: ___ matches to any subtree and automatically returns a mapping
            case 4: matches only if the exact names are the same (falls through to shallow_match_generic)
        Args:
            ins_node:
            std_node:
            id_val:
            check_meta:

        Returns:
            list of AstMap: a mapping of ins_node to std_node and possibly a symbol_table, or False if it doesn't match
        """
        name_id = ins_node.astNode.__getattribute__(id_val)
        match = _name_regex(name_id)
        mapping = AstMap()
        matched = False
        # TODO: add functionality to add function references to func_table?
        meta_matched = self.metas_match(ins_node, std_node, check_meta)
        if match[_VAR] and meta_matched:  # variable
            if type(std_node.astNode).__name__ == "Name" or id_val in ["attr", "arg"]:
                if id_val in ["attr", "arg"]:
                    std_node.astNode._id = std_node.astNode.__getattribute__(id_val)
                if std_node.field == "func" and ins_node.field != _NONE_FIELD:
                    # TODO: This 'ins_node.field != _NONE_FIELD' code is for an obscure edge case where the
                    #  instructor code is only _var_
                    std_node.astNode._id = std_node.astNode.__getattribute__(id_val)
                    mapping.add_func_to_sym_table(ins_node, std_node)
                else:
                    std_node.astNode._id = std_node.astNode.__getattribute__(id_val)
                    mapping.add_var_to_sym_table(ins_node, std_node)  # TODO: Capture result?
                matched = True
        # could else return False, but shallow_match_generic should do this as well
        elif match[_EXP] and meta_matched:
            mapping.add_exp_to_sym_table(ins_node, std_node)
            matched = True
        elif match[_WILD] and meta_matched:
            matched = True

        if matched:
            mapping.add_node_pairing(ins_node, std_node)
            return [mapping]
        # else
        return self.shallow_match_main(ins_node, std_node, check_meta=check_meta, ignores=["ctx"])

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def shallow_match_arg(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        ins_node.astNode._id = ins_node.arg
        # TODO: annotations are currently ignored because shallow_symbol_handler doesn't handle them, feature? or
        #  should we fix this. Although this should actually be toggleable?
        return self.shallow_symbol_handler(ins_node, std_node, "arg", check_meta=check_meta)

    def shallow_match_arguments(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        # TODO: do we ignore default values? Currently not ignored
        return self.shallow_match_generic(ins_node, std_node, check_meta=check_meta)

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def shallow_func_handle(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        if ins_node.field == "func" and std_node.field == "func":
            ins_node.astNode._id = ins_node.astNode.attr
            return self.shallow_symbol_handler(ins_node, std_node, "attr", check_meta)
        return self.shallow_match_generic(ins_node, std_node, check_meta)

    def shallow_match_Attribute(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        if ins_node.field == "func" and std_node.ast_name == "Attribute":
            return self.shallow_func_handle(ins_node, std_node, check_meta)
        elif std_node.ast_name == "Attribute":
            ins_node.astNode._id = ins_node.attr  # TODO: Fix this hack more gracefully
            # add_var_to_sym_table in ast_map needs the id attribute to make the map
            return self.shallow_symbol_handler(ins_node, std_node, "attr", check_meta)
        else:
            return self.shallow_match_generic(ins_node, std_node, check_meta)

    # noinspection PyPep8Naming
    def shallow_match_Name(self, ins_node, std_node, check_meta=True):
        """
        TODO: Make this handle the func field to handle functions
        Matches ins_node to std_node for different cases of encountering a name node in ins_node
            case 1: _var_ matches if std_node is a name node and automatically returns a mapping and symbol table
            case 2: __exp__ matches to any subtree and automatically returns a mapping and symbol table
            case 3: ___ matches to any subtree and automatically returns a mapping
            case 4: matches only if the exact names are the same (falls through to shallow_match_generic)
        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:
            list of AstMap: a mapping of ins_node to std_node and possibly a symbol_table, or False if it doesn't match
        """
        ins_node.ast_node._id = ins_node.id
        return self.shallow_symbol_handler(ins_node, std_node, "id", check_meta)

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def shallow_match_Pass(self, ins_node, std_node, check_meta=True):
        """
        An empty body should match to anything
        TODO: This should match to an entire body rather than just a single statement in a body
        Args:
            ins_node: Instructor ast to find in the student ast
            std_node: Student AST to search for the instructor ast in
            check_meta: flag to check whether the fields of the instructor node and the student node should match

        Returns:
            list of AstMap: a mapping between the instructor and student asts, or False if such a mapping doesn't exist
        """
        # if check_meta and ins_node.field != std_node.field:
        if not self.metas_match(ins_node, std_node, check_meta):
            return []
        mapping = AstMap()
        mapping.add_node_pairing(ins_node, std_node)
        return [mapping]

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def shallow_match_Expr(self, ins_node, std_node, check_meta=True):
        """
        An Expression node (not to be confused with expressions denoted by the instructor nodes in Name ast nodes)
        should match to anything
        Args:
            ins_node: Instructor ast to find in the student ast
            std_node: Instructor ast to find in the student ast
            check_meta: flag to check whether the fields of the instructor node and the student node should match

        Returns:
            a mapping between the instructor and student asts, or False if such a mapping doesn't exist
        """
        # if check_meta and ins_node.field != std_node.field:
        if not self.metas_match(ins_node, std_node, check_meta):
            return []
        mapping = AstMap()
        mapping.add_node_pairing(ins_node, std_node)
        return [mapping]

    def shallow_match_Call(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        return self.shallow_match_main(ins_node, std_node, check_meta, ignores=None)
        # matches = self.shallow_match_main(ins_node, std_node, check_meta, ignores=["func"])
        # if matches:
        #    pass
        # return None
        # TODO: Make this handle Calls more intelligently

    def shallow_match_xDef(self, is_match, mapping, ins_node, std_node, meta_matched, table_func):
        """
        This function is for ClassDef, FuncDef, and any similar type of ast node that has a name
        Args:
            is_match ():
            mapping ():
            ins_node ():
            std_node ():
            meta_matched ():
            table_func (str): the name of the table function to add (see ast_map)
        Returns:

        """
        ins = ins_node.astNode
        std = std_node.astNode
        matched = False
        if is_match and mapping:
            name = ins.name
            match = _name_regex(name)
            if match[_VAR] and meta_matched:  # variable
                ins._id = name
                std._id = std.name
                target_func = getattr(mapping[0], table_func, mapping[0].add_func_to_sym_table)
                target_func(ins_node, std_node)  # TODO: Capture result?
                matched = True
            elif match[_WILD] and meta_matched:
                matched = True
            elif name == std.name and meta_matched:
                matched = True
        if matched:
            return mapping
        else:
            return []

    # noinspection PyPep8Naming
    def shallow_match_FunctionDef(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        ins = ins_node.astNode
        std = std_node.astNode
        meta_matched = self.metas_match(ins_node, std_node, check_meta)
        is_match = type(ins).__name__ == type(std).__name__ and meta_matched
        mapping = self.shallow_match_main(ins_node, std_node, check_meta, ignores=['name', 'args'])
        return self.shallow_match_xDef(is_match, mapping, ins_node, std_node, meta_matched, SymTables['FUNC'])

    def shallow_match_ClassDef(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        ins = ins_node.astNode
        std = std_node.astNode
        meta_matched = self.metas_match(ins_node, std_node, check_meta)
        is_match = type(ins).__name__ == type(std).__name__ and meta_matched
        mapping = self.shallow_match_main(ins_node, std_node, check_meta, ignores=['name'])
        return self.shallow_match_xDef(is_match, mapping, ins_node, std_node, meta_matched, SymTables['CLASS'])

    # noinspection PyMethodMayBeStatic
    def shallow_match_generic(self, ins_node, std_node, check_meta=True):
        """
        Checks that all non astNode attributes are equal between ins_node and std_node
        Args:
            ins_node: Instructor ast root node
            std_node: Student AST root node
            check_meta: flag to check whether the fields of the instructor node and the student node should match

        Returns:
            list of AstMap: a mapping between the instructor and student root nodes (potentially empty)
        """
        return self.shallow_match_main(ins_node, std_node, check_meta=check_meta)

    def shallow_match_main(self, ins_node, std_node, check_meta=True, ignores=None):
        """
        Checks that all non astNode attributes are equal between ins_node and std_node
        Args:
            ins_node: Instructor ast root node
            std_node: Student AST root node
            check_meta: flag to check whether the fields of the instructor node and the student node should match
            ignores: a mapping between the instructor and student root nodes, or False if such a mapping doesn't exist

        Returns:

        """
        if ignores is None:
            ignores = []
        ignores.append("_id")  # special exception for symbols in lookup tables
        ins = ins_node.astNode
        std = std_node.astNode
        ins_field_list = list(ast.iter_fields(ins))
        std_field_list = list(ast.iter_fields(std))
        meta_matched = self.metas_match(ins_node, std_node, check_meta)
        is_match = len(ins_field_list) == len(std_field_list) and type(ins).__name__ == type(
            std).__name__ and meta_matched
        for insTup, stdTup in zip(ins_field_list, std_field_list):
            if not is_match:
                break

            ins_field = insTup[0]
            ins_value = insTup[1]
            std_field = stdTup[0]
            std_value = stdTup[1]

            if ins_value is None:
                continue

            ignore_field = ins_field in ignores

            is_match = (ins_field == std_field) or ignore_field

            if not isinstance(ins_value, list):
                ins_value = [ins_value]

            if not isinstance(std_value, list):
                std_value = [std_value]

            # is_match = len(ins_value) == len(std_value)# for stretchy matching this isn't True
            # Reference ast_node_visitor.js for the original behavior and keep note of it for the purposes of handling
            # the children noting the special case when the nodes of the array are actually parameters of the node
            # (e.g. a load function) instead of a child node
            if not ignore_field:
                for inssub_value, stdsub_value in zip(ins_value, std_value):
                    if not is_match:
                        break
                    # TODO: make this a smarter comparison, maybe handle dictionaries, f-strings, tuples, etc.
                    if is_primitive(inssub_value):
                        is_match = inssub_value == stdsub_value
        if is_match:
            mapping = AstMap()  # return MAPPING
            mapping.add_node_pairing(ins_node, std_node)
            return [mapping]
        else:
            return []

    # filter function for various types of nodes
    def shallow_match(self, ins_node, std_node, check_meta=True):
        """

        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        method_name = 'shallow_match_' + type(ins_node.astNode).__name__
        target_func = getattr(self, method_name, self.shallow_match_generic)
        return target_func(ins_node, std_node, check_meta=check_meta)

    @staticmethod
    def metas_match(ins_node, std_node, check_meta=True):
        """
        Args:
            ins_node:
            std_node:
            check_meta:

        Returns:

        """
        return ((check_meta and ins_node.field == std_node.field) or
                not check_meta
                # or std_node.field == _NONE_FIELD
                or ins_node.field == _NONE_FIELD)



import setuptools

setuptools.setup(
    name='pedal',
    version='2.6.1',
    python_requires='>=3.7',
    author='acbart,lukesg08',
    author_email='acbart@udel.edu',
    description='Tools to provide feedback on student code.',
    keywords='feedback education teaching program analysis tifa cait sandbox pedal grading grader grade',
    packages=['pedal', 'pedal.core', 'pedal.utilities', 'pedal.environments',
              'pedal.resolvers', 'pedal.command_line',
              'pedal.source', 'pedal.cait', 'pedal.tifa',
              'pedal.sandbox', 'pedal.sandbox.library', 'pedal.assertions',
              'pedal.questions', 'pedal.extensions', 'pedal.types', 'pedal.types.library'
              ],
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    long_description=open('README.rst').read(),
    install_requires=["tabulate"],
    entry_points={
          'console_scripts': [
              'pedal = pedal.command_line.command_line:main'
          ]
      },
    url='https://pedal-edu.github.io/pedal',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Education",
        "Topic :: Education",
    ]
)



# from __future__ import print_function
import unittest
import ast
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pedal.cait.stretchy_tree_matching import *
from pedal.cait.cait_node import *
from pedal.source import set_source, verify
from pedal.tifa import tifa_analysis
from pedal.core.report import MAIN_REPORT
from pedal.core.commands import clear_report, contextualize_report
from pedal.cait.cait_api import *

'''
_accu_ = 0
_iList_ = [1,2,3,4]
for _item_ in _iList_:
    accu_ = _accu_ + _item_
print(_accu_)
'''


def parse_code(student_code):
    """
    parses code and returns a new CaitNode that is the root of the student
    """
    return CaitNode(ast.parse(student_code))


class CaitTests(unittest.TestCase):

    def setUp(self):
        """

        """
        MAIN_REPORT.clear()

    def test_var_match(self):
        """
        TODO: Things about symbols we haven't asked about
                - how does scope work? Right now everything is stored
                   in one name space because symbols don't currently
                   have an idea of "scope".
        Returns:

        """
        # print("TESTING VAR MATCH")
        # tests whether variables are stored correctly
        test_tree = StretchyTreeMatcher("_accu_ = 0", report=MAIN_REPORT)

        # extracting instrutor name node
        ins_ast = test_tree.root_node
        ins_name_node = ins_ast.children[0].children[0]

        # creating student code
        student_code = "accumulator = 0"
        std_ast = parse_code(student_code)
        # extracting student name node
        std_name_node = std_ast.children[0].children[0]
        mapping = test_tree.shallow_match(ins_name_node, std_name_node)
        self.assertTrue(mapping, "match not found")

        if mapping:
            mapping = mapping[0]
            self.assertTrue(list(mapping.mappings.keys())[0].astNode ==
                            ins_name_node.astNode, "ins node match not correct in _var_ case")
            self.assertTrue(list(mapping.mappings.values())[0].astNode ==
                            std_name_node.astNode, "std node match not correct in _var_ case")
            debug_print = False  # TODO: debug flag
            if debug_print:
                print(test_tree)
                print(ins_name_node)
                print(std_name_node)
                print(mapping)

    def test_expression_match(self):
        # tests whether expressions are stored correctly
        # print("TESTING EXPRESSION MATCH")
        test_tree = StretchyTreeMatcher("__exp__ = 0", report=MAIN_REPORT)

        # extracting instructor name node
        ins_ast = test_tree.root_node
        ins_name_node = ins_ast.children[0].children[0]

        # creating student code
        student_code = "accumulator = 0\nfor item in myList:\n    accumulator += item"
        std_ast = parse_code(student_code)
        # extracting body node
        std_for_loop = std_ast.children[1]

        mapping = test_tree.shallow_match(ins_name_node, std_for_loop, False)
        self.assertTrue(mapping, "match not found")
        if mapping:
            mapping = mapping[0]
            if mapping:
                keys = list(mapping.exp_table.keys())
                values = list(mapping.exp_table.values())
                self.assertTrue(keys[0] == "__exp__", "symbol match not found")
                self.assertTrue(values[0].astNode == std_for_loop.astNode,
                                "did not match to correct ast node")
            debug_print = False  # TODO: debug flag
            if debug_print:
                print(test_tree)
                print(ins_name_node)
                print(std_for_loop)
                print(mapping)

        student_code2 = ("for item in item_list:\n"
                         "    if item > 5:\n"
                         "        item_sum = item_sum + item")
        matcher1 = StretchyTreeMatcher("for ___ in ___:\n"
                                       "    __expr__", report=MAIN_REPORT)
        matches02 = matcher1.find_matches(student_code2)
        self.assertTrue(matches02, "Expression match doesn't match to subtree")

    def test_function_diving(self):
        # TODO: Fix this bug
        student_code1 = "print(report['station'])"
        student_code2 = "print(report['station']['city'])"

        matcher1 = StretchyTreeMatcher("___[__expr__]", report=MAIN_REPORT)

        matches01 = matcher1.find_matches(student_code1)
        matches02 = matcher1.find_matches(student_code2)

        self.assertTrue(matches01)
        self.assertTrue(matches02)

    def test_function_save(self):
        ins_code = ("def _my_func_():\n"
                    "   pass\n"
                    "_my_func_()")
        std_code = ("def funky(forum):\n"
                    "    pass\n"
                    "funky(1200)")  # matches

        std_code2 = ("def funky(forum):\n"
                     "    pass\n"
                     "dunk(1200)")  # matches

        matcher = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        matches = matcher.find_matches(std_code)

        self.assertTrue(matches, "did not match")
        table = matches[0].func_table
        keys = list(table.keys())
        self.assertTrue(len(keys) == 1, "found {} keys, should have found 1".format(len(keys)))
        self.assertTrue(keys[0] == '_my_func_',
                        "expected function name to be '_my_func_', found {} instead".format(keys[0]))
        self.assertTrue(len(table['_my_func_']) == 2, "function call and definition not counted as separate instances")

        matches2 = matcher.find_matches(std_code2)
        self.assertTrue(len(matches2) == 0, "Function calls and names aren't correlating")

    def test_wild_card_match(self):
        # tests whether Wild matches are stored correctly
        # print("TESTING WILD CARD MATCH")
        test_tree = StretchyTreeMatcher("___ = 0", report=MAIN_REPORT)

        # extracting instrutor name node
        ins_ast = test_tree.root_node
        ins_name_node = ins_ast.children[0].children[0]

        # creating student code
        student_code = "accumulator = 0\nfor item in myList:\n    accumulator += item\nfor item in myList:" \
                       "\n    accumulator += item"
        std_ast = parse_code(student_code)
        # extracting body node
        std_for_loop = std_ast.children[1]
        mapping = test_tree.shallow_match(ins_name_node, std_for_loop, False)
        self.assertTrue(mapping, "match not found")
        if mapping:
            mapping = mapping[0]
            self.assertTrue(list(mapping.mappings.keys())[0].astNode == ins_name_node.astNode and
                            list(mapping.mappings.values())[0].astNode == std_for_loop.astNode,
                            "wild card didn't match")
            debug_print = False  # TODO: debug flag
            if debug_print:
                print(test_tree)
                print(ins_name_node)
                print(std_for_loop)
                print(mapping)

    def test_match_all(self):
        # print("TESTING MATCH ALL")
        test_tree = StretchyTreeMatcher("___ = 0", report=MAIN_REPORT)
        ins_ast = test_tree.root_node
        ins_num_node = ins_ast.children[0].children[1]

        # creating student code
        student_code = "billy = 0"
        std_ast = parse_code(student_code)
        # extracting body node
        std_num_node = std_ast.children[0].children[1]

        mapping = test_tree.shallow_match(ins_num_node, std_num_node)
        self.assertTrue(mapping, "match not found")
        if mapping:
            mapping = mapping[0]
            if mapping:
                self.assertTrue(list(mapping.mappings.keys())[0].astNode == ins_num_node.astNode,
                                "ins node not matched correctly")
                self.assertTrue(list(mapping.mappings.values())[0].astNode == std_num_node.astNode,
                                "student node not matched correctly")
            debug_print = False  # TODO: debug print
            if debug_print:
                print(test_tree)
                print(std_num_node)
                print(mapping)

    def test_generic_match(self):
        # print("TESTING SAME ORDER")
        std_code = ("_sum = 0\n"
                    "_list = [1,2,3,4]\n"
                    "for item in _list:\n"
                    "    _sum = _sum + item\n"
                    "print(_sum)")
        ins_code = ('_accu_ = 0\n'
                    '_iList_ = __listInit__\n'
                    'for _item_ in _iList_:\n'
                    '    _accu_ = _accu_ + _item_\n'
                    'print(_accu_)')
        # var std_code = "_sum = 12 + 13"
        # var ins_code = "_accu_ = 12 + 11"
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)

        mappings_array = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(mappings_array, "matches not found")
        if mappings_array:
            mappings = mappings_array[0]
            self.assertTrue(len(mappings["_accu_"]) == 4, "Incorrect variable map")
            self.assertTrue(mappings["_accu_"].id == "_sum", "Incorrect variable map")

            self.assertTrue(len(mappings["_iList_"]) == 2, "Incorrect variable map")
            self.assertTrue(mappings["_iList_"].id == "_list", "Incorrect variable map")

            self.assertTrue(len(mappings["_item_"]) == 2, "Incorrect variable map")
            self.assertTrue(mappings["_item_"].id == "item", "Incorrect variable map")
            # _accu_/_sum is 4
            # _iList_/_list is 2
            # _item_/item is 2
            for k, v in mappings.mappings.items():
                try:
                    if k.id == "__lisetInit__":
                        self.assertTrue("List" == v.ast_name,
                                        "invalid mapping, {} mapped to {}".format(k.ast_name, v.ast_name))
                except AttributeError:
                    self.assertTrue(k.ast_name == v.ast_name,
                                    "invalid mapping, {} mapped to {}".format(k.ast_name, v.ast_name))
            debug_print = False  # TODO: debug print
            if debug_print:
                print(mappings)
                print(ins_ast.astNode)
                print(std_ast.astNode)

    def test_commutativity(self):
        # print("TESTING COMMUTATIVITY ADDITION")
        # fail_count = 0
        # success_count = 0
        std_code = ("_sum = 0\n"
                    "list = [1,2,3,4]\n"
                    "for item in list:\n"
                    "    _sum = item + _sum\n"
                    "print(_sum)")
        ins_code = ("_accu_ = 0\n"
                    "_iList_ = __listInit__\n"
                    "for _item_ in _iList_:\n"
                    "    _accu_ = _accu_ + _item_\n"
                    "print(_accu_)")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)

        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(mappings, "mapping not found")
        if mappings:
            mappings = mappings[0]
            self.assertTrue(mappings.conflict_keys != 0, "Conflicting keys in ADDITION when there shouldn't be")
            debug_print = False  # TODO: debug print
            if debug_print:
                print(ins_ast.astNode)
                print(std_ast.astNode)
                print(mappings)

    def test_many_to_one(self):
        # print("TESTING MANY TO ONE")
        std_code = ("_sum = 0\nlist = [1,2,3,4]\n"
                    "for item in list:\n"
                    "    _sum = _sum + item\n"
                    "print(_sum)")
        ins_code = ("_accu_ = 0\n"
                    "_iList_ = __listInit__\n"
                    "for _item_ in _iList_:\n"
                    "    _accu_ = _accu2_ + _item_\n"
                    "print(_accu_)")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)

        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(mappings, "no matches found")
        if mappings:
            mappings = mappings[0]
            if mappings:
                self.assertTrue(len(mappings.conflict_keys) == 0, "Conflicting keys when there shouldn't be")
                self.assertEqual(len(mappings.symbol_table), 4, "inconsistent symbol matching")
                lengths = sorted([len(v) for v in mappings.symbol_table.values()])
                self.assertEqual(lengths, [1, 2, 2, 3], "inconsistent symbol matching")
            debug_print = False
            if debug_print:
                print(ins_ast.astNode)
                print(std_ast.astNode)
                print(mappings)

    def test_one_to_many(self):
        # print("TESTING ONE TO MANY")
        std_code = ("_sum = 0\n"
                    "list = [1,2,3,4]\n"
                    "for item in list:\n"
                    "    _sum = _sum + _sum\n"
                    "print(_sum)")
        ins_code = ("_accu_ = 0\n"
                    "_iList_ = __listInit__\n"
                    "for _item_ in _iList_:\n"
                    "    _accu_ = _accu_ + _item_\n"
                    "print(_accu_)")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)
        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertFalse(mappings, "found match when match shouldn't be found")
        debug_print = False  # TODO: debug print
        if debug_print:
            print(ins_ast.astNode)
            print(std_ast.astNode)
            print(mappings)

    def test_multimatch_false(self):
        # print("TESTING MULTI-MATCH")
        # var std_code = "_sum = 0\ncount = 0\nlist = [1,2,3,4]\nfor item in list:\n    _sum = _sum + item\n    count" \
        # " = count + 1\nprint(_sum)"
        std_code = ("_sum = 0\n"
                    "count = 0\n"
                    "_list = [1,2,3,4]\n"
                    "for item in _list:\n"
                    "    _sum = _sum + count\n"
                    "    count = _sum + count\n"
                    "print(_sum) ")
        ins_code = ("_accu_ = 0\n"
                    "_iList_ = __listInit__\n"
                    "for _item_ in _iList_:\n"
                    "    _accu_ = _accu_ + __exp__\n"
                    "print(_accu_)")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)
        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertFalse(len(mappings) > 1, "too many matches found")

        debug_print = False
        if debug_print:
            print(ins_ast)
            print(std_ast)
            print(mappings)

    def test_multimatch_true(self):
        std_code = ("_sum = 0\n"
                    "count = 0\n"
                    "_list = [1,2,3,4]\n"
                    "for item in _list:\n"
                    "    _sum = _sum + count\n"
                    "    count = _sum "
                    "+ count\n"
                    "print(_sum) ")
        ins_code = ("_accu_ = 0\n"
                    "_iList_ = __listInit__\n"
                    "for _item_ in _iList_:\n"
                    "    _accu_ = _accu_ + __exp__")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)
        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(len(mappings) >= 2, "Not enough mappings found")
        debug_print = False
        if debug_print:
            print(ins_ast)
            print(std_ast)
            print(mappings)

    def test_pass(self):
        # print("TESTING PASS MATCH")
        std_code = ('import matplotlib.pyplot as plt\n'
                    'quakes = [1,2,3,4]\n'
                    'quakes_in_miles = []\n'
                    'for quakes in quakes:\n'
                    '    quakes_in_miles.append(quake * 0.62)\n'
                    'plt.hist(quakes_in_miles)\n'
                    'plt.xlabel("Depth in Miles")\n'
                    'plt.ylabel("Number of Earthquakes")\n'
                    'plt.title("Distribution of Depth in Miles of Earthquakes")\n'
                    'plt.show()')
        ins_code = ("for _item_ in _item_:\n"
                    "    pass")
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)
        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(mappings, "mapping should have been found")
        debug_print = False
        if debug_print:
            print(ins_ast)
            print(std_ast)
            print(mappings)

    def test_meta_stretch(self):
        # print("TESTING META STRETCH")
        std_code = ''.join([
            'steps_hiked_list = [1,2,3,4]\n',
            'total = 0\n',
            'for steps_hiked in steps_hiked_list:\n',
            '    total = steps_hiked + total\n',
            '    steps = steps + 1'])
        ins_code = "for ___ in ___:\n" \
                   "    ___ = _sum_ + ___"
        ins_tree = StretchyTreeMatcher(ins_code, report=MAIN_REPORT)
        ins_ast = ins_tree.root_node
        std_ast = parse_code(std_code)
        mappings = ins_tree.find_matches(std_ast.astNode)
        self.assertTrue(mappings, "mapping should have been found")
        if mappings:
            self.assertTrue(len(mappings) == 3,
                            "Should find exactly 3 matches, instead found {count}".format(
                                count=len(mappings).__str__()))
        debug_print = False
        if debug_print:
            print(ins_ast)
            print(std_ast)
            print(mappings)

    def test_parse_program(self):
        # noinspection PyBroadException
        try:
            parse_program()
            self.assertTrue(False, "Program did not throw exception")
        except Exception:
            self.assertTrue(True, "Should NOT FAIL")
        contextualize_report("fun = 1 + 0")
        parse_program()
        self.assertTrue('cait' in MAIN_REPORT, "No parsing happened")

    """
    def test_def_use_error(self):
        contextualize_report("fun = fun + 1")
        parse_program()
        name_node = MAIN_REPORT["source"]["ast"].body[0].cait_node.target
        self.assertTrue(def_use_error(name_node), "def_use error should have been found but wasn't")
        self.assertTrue(def_use_error("fun"), "def_use error should have been found but wasn't")
        self.assertFalse(def_use_error("gitax"), "variable doesn't exist")

    def test_data_type(self):
        contextualize_report("fun = 1 + 1")
        parse_program()
        name_node = MAIN_REPORT["source"]["ast"].body[0].cait_node.target
        self.assertTrue(data_type(name_node).is_equal(int), "Data type not successfully found from name node")
        self.assertTrue(data_type("fun").is_equal(int), "Data type not successfully found from str name")"""

    def test_find_match(self):
        contextualize_report("fun = 1 + 1")
        parse_program()
        match = find_match("_var_ = __expr__")
        self.assertIsInstance(match, AstMap, "Match not found")

    def test_find_matches(self):
        contextualize_report("fun = 1 + 1\nfun2 = 2 + 2")
        parse_program()
        matches = find_matches("_var_ = __expr__")
        self.assertIsInstance(matches, list, "find_matches did not return a list")
        self.assertIsInstance(matches[0], AstMap, "find_matches does not contain an AstMap")
        self.assertEqual(len(matches), 2, "find_matches does not return the correct number of matches")

        contextualize_report("for 'fun' in ___:"
                             "if ___:\n"
                             "    pass")
        matches = find_matches("for ___ in ___:\n"
                               "    pass")
        self.assertFalse(matches, "find matches should have misparsed and returned False, but returned True instead")

    def test_invalid_code(self):
        contextualize_report("float('0') + 1")
        parse_program()
        matches = find_match("_var_ = __expr__")
        self.assertIsNone(matches)

    def test_old_style_api(self):
        contextualize_report("a = open('file.txt')")
        std_ast = parse_program()
        calls = std_ast.find_all("Call")
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0].func.ast_name, 'Name')
        self.assertEqual(calls[0].func.id, 'open')
        self.assertEqual(len(calls[0].args), 1)

        clear_report()
        contextualize_report("def a():\n  pass\na()")
        std_ast = parse_program()
        defs = std_ast.find_all("FunctionDef")
        self.assertEqual(len(defs), 1)

        clear_report()
        contextualize_report("1 < 1")
        std_ast = parse_program()
        compares = std_ast.find_all("Compare")
        self.assertEqual(len(compares), 1)

        clear_report()
        contextualize_report("for x in y:\n  x")
        std_ast = parse_program()
        loops = std_ast.find_all("For")
        self.assertEqual(len(loops), 1)
        self.assertEqual(loops[0].target.ast_name, "Name")

        # Multiple assignment
        clear_report()
        contextualize_report("a, b = 0, 1")
        std_ast = parse_program()
        assigns = std_ast.find_all("Assign")
        self.assertEqual(len(assigns), 1)
        self.assertEqual(len(assigns[0].targets), 1)
        self.assertEqual(assigns[0].targets[0].ast_name, 'Tuple')
        self.assertEqual(assigns[0].targets[0].elts[0].id, 'a')

        clear_report()
        contextualize_report('from pprint import *')
        parse_program()

    def test_matches_in_matches(self):
        matcher1 = StretchyTreeMatcher("if __expr__:\n"
                                       "    pass", report=MAIN_REPORT)
        matcher2 = StretchyTreeMatcher("0.4*_item_", report=MAIN_REPORT)
        student_code = ("if 0.4*item < 40:\n"
                        "    pass")
        self.assertTrue(matcher2.find_matches(student_code, check_meta=False))
        matches = matcher1.find_matches(student_code)
        self.assertTrue(matches)

        __expr__ = matches[0]["__expr__"]
        matches2 = __expr__.find_matches("0.4*_item_", )
        self.assertTrue(matches2)

        __expr__ = matches[0]["__expr__"]
        matches3 = __expr__.find_matches(__expr__, )
        self.assertTrue(matches3)

    def test_symbol_mapping(self):
        student_code = ("item = 0\n"
                        "item2 = 1")
        matcher1 = StretchyTreeMatcher("_item_ = 0\n"
                                       "_item_ = 1", report=MAIN_REPORT)
        match = matcher1.find_matches(student_code)
        self.assertFalse(match, "One student variable is mapping to multiple instructor variables")

        student_code = ("item = 0\n"
                        "item = 1")
        matcher2 = StretchyTreeMatcher("_item_ = 0\n"
                                       "_item2_ = 1", report=MAIN_REPORT)
        match2 = matcher2.find_matches(student_code)
        self.assertTrue(match2,
                        "One instructor variable should be able to map to multiple student variables, but isn't")

    def test_tree_cutting(self):
        student_code1 = ('maxtemp_list = []\n'
                         'mintemp_list = []\n'
                         'for report in weather_reports:\n'
                         '    if report["Station"]["City"] == "Blacksburg":\n'
                         '        maxtemp_list.append(report["Data"]["Temperature"]["Max Temp"])\n'
                         '        mintemp_list.append(report["Data"]["Temperature"]["Min Temp"])')

        student_parse = CaitNode(ast.parse(student_code1))
        matcher1 = StretchyTreeMatcher("'Blacksburg'", report=MAIN_REPORT)
        res = find_expr_sub_matches("'Blacksburg'", student_parse)
        self.assertTrue(res, "Cutting broke sub-matching")

        matches01 = matcher1.find_matches(student_code1)
        self.assertTrue(matches01, "Cutting doesn't work")

        student_code2 = ("def my_func(funky):\n"
                         "    print(funky)")
        matcher2 = StretchyTreeMatcher("def _func_def_():\n"
                                       "    pass", report=MAIN_REPORT)
        matches02 = matcher2.find_matches(student_code2)
        self.assertTrue(matches02, "Function reserved doesn't work with cutting")
        self.assertTrue(len(matches02[0].func_table.keys()) == 1, "Function node is ignored")

        student_code3 = "x"
        matcher3 = StretchyTreeMatcher("_x_", report=MAIN_REPORT)
        matches03 = matcher3.find_matches(student_code3)
        self.assertTrue(matches03, "Function reserved doesn't work with cutting")

    def test___getitem__(self):
        student_code2 = ("def my_func(funky):\n"
                         "    print(funky)")
        matcher2 = StretchyTreeMatcher("def _func_def_():\n"
                                       "    print(_funky_)", report=MAIN_REPORT)
        matches02 = matcher2.find_matches(student_code2)
        self.assertTrue(matches02, "match not found, aborting test")
        self.assertTrue(len(matches02[0].func_table.keys()) == 1, "improper number of keys, aborting test")
        self.assertTrue(matches02[0]['_func_def_'], "Couldn't retrieve function name.")
        self.assertTrue(matches02[0]['_funky_'], "Couldn't retrieve variable name.")

    @unittest.skip("Does not support keyword arguments")
    def test_function_keyarg_matches(self):
        """
        TODO: This has a lot of aspects, such as
                - are keywords symbols that should be saved?
                - how are keywords stored in the name space?
                - how does that interact with the name space stuff
                    in variable matches?
        Returns:

        """
        self.assertTrue(False)

    def test_function_arg_matches(self):
        student_code2 = ("def my_func(funky):\n"
                         "    print(funky)")
        matcher2 = StretchyTreeMatcher("def _func_def_(_funky_):\n"
                                       "    pass", report=MAIN_REPORT)
        matches02 = matcher2.find_matches(student_code2)
        match02 = matches02[0]
        self.assertTrue(matches02, "match not found, aborting test")
        self.assertTrue(len(match02.func_table.keys()) == 1, "improper number of keys, aborting test")
        self.assertTrue(match02['_func_def_'], "Couldn't retrieve function name.")
        self.assertTrue(match02['_funky_'], "Couldn't retrieve variable name.")
        self.assertTrue(match02['_funky_'].id == "funky",
                        "Incorrect symbol name, expected 'funky' but got '{}' instead".format(match02['_funky_'].id))

        student_code3 = ("def my_func(funky, dunky):\n"
                         "    print(funky + dunky)")
        matcher3 = StretchyTreeMatcher("def _func_def_(_funky_, _funky2_):\n"
                                       "    pass", report=MAIN_REPORT)
        matches03 = matcher3.find_matches(student_code3)
        match03 = matches03[0]
        self.assertTrue(matches03, "match not found, aborting test")
        self.assertTrue(len(match03.func_table.keys()) == 1, "improper number of keys, aborting test")
        self.assertTrue(match03['_func_def_'], "Couldn't retrieve function name.")
        self.assertTrue(match03['_funky_'], "Couldn't retrieve variable name.")
        self.assertTrue(match03['_funky_'].id == "funky",
                        "Incorrect symbol name, expected 'funky' but got '{}' instead".format(match02['_funky_'].id))
        self.assertTrue(match03['_funky2_'], "Couldn't retrieve variable name.")
        self.assertTrue(match03['_funky2_'].id == "dunky",
                        "Incorrect symbol name, expected 'dunky' but got '{}' instead".format(match03['_funky_'].id))

    def test_cait_node_is_method(self):
        contextualize_report("class Dog:\n def bark(self):\n  pass\ndef dogs_bark(alod):\n pass")
        ast = parse_program()
        functions = ast.find_all('FunctionDef')
        self.assertTrue(functions[0].is_method())
        self.assertFalse(functions[1].is_method())
        classes = ast.find_all('ClassDef')
        self.assertFalse(classes[0].is_method())
        passes = ast.find_all('Pass')
        self.assertFalse(passes[0].is_method())
        self.assertFalse(passes[1].is_method())

    def test_ast_map_symbols(self):
        contextualize_report("fun = 0\nfun = 1")
        parse_program()
        my_match = find_match("_fun_ = ___")
        self.assertTrue(my_match["_fun_"].id == "fun")
        self.assertTrue(my_match["_fun_"][0].id == "fun")

    def test_function_Names(self):
        contextualize_report("import weather\n"
                             "weather_reports = get_weather('Data')\n"
                             "weather_reports = weather.get_weather('Data')\n"
                             "total = 0\n"
                             "for weather in weather_reports:\n"
                             "    total = total + weather\n"
                             "print(total)")
        parse_program()
        my_match = find_matches("_get_weather_(__expr__)")
        #  TODO: add to func table instead of symbol table?
        self.assertTrue(my_match[0]['_get_weather_'].id == "get_weather", "Function wasn't properly retrieved")
        self.assertTrue(my_match[1]['_get_weather_'].id == "print")

        my_match2 = find_matches("_var_._method_()")
        self.assertTrue(len(my_match2) == 1, "couldn't find attr func access")

        my_match3 = find_matches("weather._method_()")
        self.assertTrue(len(my_match3) == 1, "couldn't find attr func access")

        my_match4 = find_matches("_var_.get_weather()")
        self.assertTrue(len(my_match4) == 1, "couldn't find attr func access")

        MAIN_REPORT.clear()
        contextualize_report('book = {"number_of_pages":285, "price":99.23, "discount":0.1}\n'
                             'print(["price"])')
        parse_program()
        match = find_match("_var_([__str1__])")
        self.assertTrue(match["_var_"].id == "print", "couldn't find attr func access")

        contextualize_report('print("fun")')
        parse_program()
        matches = find_matches("_var_")
        var = matches[0]["_var_"]
        self.assertTrue(var.ast_name == "Name", "is: '{}' instead of Name".format(var.ast_name))

    def test_class_Names(self):
        contextualize_report("class myClass:\n"
                             "    def f(self):\n"
                             "        return 'hello world'")
        parse_program()
        matches = find_matches("class myClass:\n\tpass")
        matches_wildcard = find_matches("class _name_:\n\tpass")

        self.assertTrue(len(matches) == 1, "exact class name matching not working")
        self.assertTrue(len(matches_wildcard) == 1, "wild card class name matching not working")
        # Added functionality for matching class names
        var = matches_wildcard[0]["_name_"]
        self.assertTrue(var.id == "myClass", "is: '{}' instead of Name".format(var.id))

    def test_attribute_names(self):
        contextualize_report("import weather\n"
                             "weather_reports = get_weather('Data')\n"
                             "weather_reports = weather.get_weather['Data']\n"
                             "total = 0\n"
                             "for weather in weather_reports:\n"
                             "    total = total + weather\n"
                             "print(total)")

        my_match1 = find_matches("_var_._attr_[__str__]")
        self.assertTrue(len(my_match1) == 1, "couldn't find attr func access")

        self.assertTrue(my_match1[0]['_attr_'].id == "get_weather", "'{}' instead of 'get_weather'".format(
            my_match1[0]['_attr_'].id))

        my_match2 = find_matches("_var_.___[__str__]")
        self.assertTrue(len(my_match2) == 1, "couldn't wild card attribute access")

    def test_broken_traversal(self):
        contextualize_report("user entered gobblydegook")
        verify()
        parse = parse_program()
        self.assertFalse(parse.find_all('For'))

    def test_aug_assignment(self):
        contextualize_report(
            "magnitudes = [1.5, 1.9, 4.3, 2.1, 2.0, 3.6, 0.5, 2.5, 1.9, 4.0, 3.8, 0.7, 2.2, 5.1, 1.6]\n"
            "count = 0\n"
            "for item in magnitudes:\n"
            "    if item > 2.0:\n"
            "        count = count + 1\n"
            "print(count)")
        match = find_matches("for ___ in ___:\n"
                             "    __exp__\n"
                             "print(_var_)")[0]
        self.assertTrue(match["__exp__"].find_matches(match["_var_"].id))

        contextualize_report(
            "magnitudes = [1.5, 1.9, 4.3, 2.1, 2.0, 3.6, 0.5, 2.5, 1.9, 4.0, 3.8, 0.7, 2.2, 5.1, 1.6]\n"
            "count = 0\n"
            "for item in magnitudes:\n"
            "    if item > 2.0:\n"
            "        count += 1\n"
            "print(count)")
        match = find_matches("for ___ in ___:\n"
                             "    __exp__\n"
                             "print(_var_)")[0]
        m__exp__ = match["__exp__"]
        submatches = m__exp__.find_matches(match["_var_"].id, check_meta=False)
        self.assertTrue(submatches)

    def test_function_recursion(self):
        contextualize_report("def recursive(item):\n"
                             "    if item < 0:\n"
                             "        return 0\n"
                             "    else:"
                             "        return recursive(item - 1)")
        matches = find_matches("def _recur_(_var_):\n"
                               "    __exp__")
        for match in matches:
            __exp__ = match["__exp__"]
            _recur_ = match["_recur_"]
            _var_ = match["_var_"]
            self.assertTrue(__exp__.find_matches("{func}()".format(func=_recur_.id)),
                            "couldn't find recursive function call")

    def test_use_previous(self):
        contextualize_report('for reports in weather_reports:\n'
                             '    if report["Station"]["City"] == "Chicago":\n'
                             '        trend.append(reports["Data"]["Precipitation"])')
        matches = find_matches("for _var_ in ___:\n"
                               "    if __expr__ == __str2__:\n"
                               "        pass")
        for match in matches:
            __expr__ = match["__expr__"]
            submatch = __expr__.find_matches("_var2_[__expr__]")
            self.assertTrue(submatch)

        contextualize_report('for reports in weather_reports:\n'
                             '    if report["Station"]["City"] == "Chicago":\n'
                             '        trend.append(reports["Data"]["Precipitation"])')
        matches = find_matches("for _var_ in ___:\n"
                               "    if __expr__ == __str2__:\n"
                               "        pass")
        for match in matches:
            __expr__ = match["__expr__"]
            submatch = __expr__.find_matches("_var_[__expr__]", use_previous=True)
            self.assertFalse(submatch)

        contextualize_report('for reports in weather_reports:\n'
                             '    if report["Station"]["City"] == "Chicago":\n'
                             '        trend.append(reports["Data"]["Precipitation"])')
        matches = find_matches("for _var_ in ___:\n"
                               "    if __expr__ == __str2__:\n"
                               "        pass")
        for match in matches:
            __expr__ = match["__expr__"]
            submatch = __expr__.find_matches("_var_[__expr__]")
            self.assertFalse(submatch)

        contextualize_report("for item in item_list:\n"
                             "    if item < 0:\n"
                             "        n = n + item")
        matches = find_matches("for _var_ in ___:\n"
                               "    __expr__")
        for match in matches:
            __expr__ = match["__expr__"]
            submatch = __expr__.find_matches("_sum_ = _sum_ + _var_", use_previous=True)
            self.assertTrue(submatch)

    def test_match_root(self):
        contextualize_report("for item in item_list:\n"
                             "    if item < 0:\n"
                             "        n = n + item\n"
                             "n = 0")
        match01 = find_match("for ___ in ___:\n"
                             "    __expr__")
        match02 = find_match("_var_ = 0")
        match03 = find_match("_var_ = _var_ + _item_")
        root01 = match01.match_root.tree_id
        root02 = match02.match_root.tree_id
        root03 = match03.match_root.tree_id
        self.assertTrue(root01 < root02)
        self.assertTrue(root01 < root03)

    def test_copy(self):
        contextualize_report("for item in item_list:\n"
                             "    if item < 0:\n"
                             "        n = n + item\n"
                             "        i = i + 1\n"
                             "n = 0")
        match01 = find_match("for _item_ in ___:\n"
                             "    __expr__")
        map_size01 = len(match01.mappings.keys())

        match02 = find_match("___ = _item_ + 1", use_previous=match01)
        map_size02 = len(match01.mappings.keys())

        match03 = find_match("___ = _item_ + ___", use_previous=match01)
        map_size03 = len(match01.mappings.keys())
        map_size03_2 = len(match03.mappings.keys())

        match04 = find_match("___ = _item_ + 1")
        match05 = find_match("___ = _item_ + ___")
        map_size05 = len(match05.mappings.keys())

        self.assertTrue(map_size01 == map_size02 == map_size03)
        self.assertNotEqual(map_size03, map_size03_2)
        self.assertNotEqual(map_size03_2, map_size05)

        self.assertFalse(match02)
        self.assertTrue(match03)
        self.assertTrue(match04)

    def test_cait_get_match_names(self):
        contextualize_report("for item in item_list:\n"
                             "    if item < 0:\n"
                             "        n = n + item\n"
                             "        i = i + 1\n"
                             "n = 0")
        match = find_match("___ = _item_ + 1")
        self.assertEqual(match.names(), {'_item_': 'i'})

    def test_find_matches_expr2(self):
        contextualize_report('for prop_report in crime_reports:\n'
                             '    total_prop= total_prop + prop_report["Property Crime"]\n'
                             '    count_prop= count_prop +1\n'
                             'average_property = total_prop/count_prop\n'
                             'for violent_report in crime_reports:\n'
                             '    total_viol= total_viol + violent_report["Violent Crime"]\n'
                             '    count_viol= count_viol +1\n'
                             'average_violent= total_viol/count_viol\n')
        matches = find_matches("for ___ in __exp__:\n"
                               "    pass")
        self.assertEqual(4, len(matches))
        # TODO: This is 4 because right now pass matches to statements instead of an entire body


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Type definitions of builtin functions and modules.
"""

from pedal.types.new_types import (make_dataclass, FunctionType, AnyType,
                                   ImpossibleType, NumType, IntType, FloatType, NoneType, BoolType,
                                   TupleType, ListType, StrType, FileType, DictType,
                                   ModuleType, SetType, LiteralInt, LiteralFloat, LiteralStr, LiteralBool,
                                   register_builtin_module,
                                   BUILTIN_MODULES, BUILTIN_NAMES,
                                   int_function, float_function, num_function,
                                   bool_function, void_function, TYPE_TYPE, exception_function,
                                   IntConstructor, FloatConstructor, BoolConstructor,
                                   ListConstructor, StrConstructor, str_function, GeneratorType, SetConstructor,
                                   FrozenSetConstructor, DictConstructor, TupleConstructor)

register_builtin_module('dataclasses', lambda: ModuleType('dataclasses', fields={
    'dataclass': FunctionType('dataclass', definition=make_dataclass)
}))

register_builtin_module('pprint', lambda: ModuleType('pprint', fields={
    'pprint': FunctionType(name='pprint', returns=NoneType)
}))

register_builtin_module('json', lambda: ModuleType('json', fields={
    'loads': FunctionType(name='loads', returns=AnyType),
    'dumps': FunctionType(name='dumps', returns=StrType)
}))

register_builtin_module('random', lambda: ModuleType('random', fields={
    'randint': FunctionType(name='randint', returns=NumType)
}))

register_builtin_module('string', lambda: ModuleType('string', fields={
    'letters': StrType(False),
    'digits': StrType(False),
    'ascii_letters': StrType(False),
    'punctuation': StrType(False),
    'printable': StrType(False),
    'whitespace': StrType(False),
    'ascii_uppercase': StrType(False),
    'ascii_lowercase': StrType(False),
    'hexdigits': StrType(False),
    'octdigits': StrType(False),
}))


def reload_math_module():
    math_module = ModuleType('math', fields={
        f.name: f for f in [
            int_function('ceil'),
            int_function('comb'),
            float_function('copysign'),
            num_function('fabs'),
            int_function('factorial'),
            int_function('floor'),
            num_function('fmod'),
            FunctionType('frexp', returns=lambda: TupleType([FloatType(), IntType()])),
            float_function('fsum'),
            int_function('gcd'),
            bool_function('isclose'),
            bool_function('isfinite'),
            bool_function('isinf'),
            bool_function('isnan'),
            int_function('isqrt'),
            int_function('lcm'),
            num_function('ldexp'),
            FunctionType('modf', returns=lambda: TupleType([FloatType(), FloatType()])),
            float_function('nextafter'),
            int_function('perm'),
            num_function('prod'),
            float_function('remainder'),
            int_function('trunc'),
            float_function('ulp'),
            float_function('exp'),
            float_function('expm1'),
            float_function('log'),
            float_function('log1p'),
            float_function('log2'),
            float_function('log10'),
            float_function('pow'),
            float_function('sqrt'),
            float_function('acos'),
            float_function('asin'),
            float_function('atan'),
            float_function('atan2'),
            float_function('cos'),
            float_function('dist'),
            float_function('hypot'),
            float_function('sin'),
            float_function('tan'),
            num_function('radians'),
            num_function('degrees'),
            float_function('acosh'),
            float_function('asinh'),
            float_function('atanh'),
            float_function('cosh'),
            float_function('sinh'),
            float_function('tanh'),
            float_function('erf'),
            float_function('erfc'),
            float_function('gamma'),
            float_function('lgamma')]
    })
    math_module.fields.update({
        'pi': FloatType(),
        'e': FloatType(),
        'tau': FloatType(),
        'inf': FloatType(),
        'nan': FloatType()
    })
    return math_module


register_builtin_module('math', reload_math_module)


def round_definition(tifa, function, callee, arguments, named_arguments, location):
    if len(arguments) == 1:
        return IntType()
    elif len(arguments) == 2:
        return FloatType()
    return ImpossibleType()


def map_definition(tifa, function, callee, arguments, named_arguments, location):
    if arguments:
        return GeneratorType(arguments[0].is_empty, arguments[0].iterate())
    return ImpossibleType()


def zip_definition(tifa, function, callee, arguments, named_arguments, location):
    tupled_types = TupleType((t.iterate() for t in arguments))
    return GeneratorType(arguments and arguments[0].is_empty,
                         tupled_types)


def enumerate_definition(tifa, function, callee, arguments, named_arguments, location):
    if arguments:
        tupled_types = TupleType((IntType(), arguments[0].iterate()))
        return GeneratorType(arguments and arguments[0].is_empty, tupled_types)
    return ImpossibleType()


# TODO: Type guard with `isinstance`
# TODO: all the built-in exceptions!

BUILT_EXCEPTION_NAMES = ["BaseException", "BaseExceptionGroup", "GeneratorExit", "KeyboardInterrupt", "SystemExit",
                         "Exception", "ArithmeticError", "FloatingPointError", "OverflowError", "ZeroDivisionError",
                         "AssertionError", "AttributeError", "BufferError", "EOFError", "ImportError",
                         "ModuleNotFoundError", "LookupError", "IndexError", "KeyError", "MemoryError", "NameError",
                         "UnboundLocalError", "OSError", "BlockingIOError", "ChildProcessError", "ConnectionError",
                         "BrokenPipeError", "ConnectionAbortedError", "ConnectionRefusedError", "ConnectionResetError",
                         "FileExistsError", "FileNotFoundError", "InterruptedError", "IsADirectoryError",
                         "NotADirectoryError", "PermissionError", "ProcessLookupError", "TimeoutError",
                         "ReferenceError", "RuntimeError", "NotImplementedError", "RecursionError",
                         "StopAsyncIteration", "StopIteration", "SyntaxError", "IndentationError", "TabError",
                         "SystemError", "TypeError", "ValueError", "UnicodeError", "UnicodeDecodeError",
                         "UnicodeEncodeError", "UnicodeTranslateError", "Warning", "BytesWarning", "DeprecationWarning",
                         "EncodingWarning", "FutureWarning", "ImportWarning", "PendingDeprecationWarning",
                         "ResourceWarning", "RuntimeWarning", "SyntaxWarning", "UnicodeWarning", "UserWarning"]

BUILTIN_NAMES.update({
    name: exception_function(name) for name in BUILT_EXCEPTION_NAMES
})

BUILTIN_NAMES.update({
    f.name: f for f in [
        TYPE_TYPE,
        IntConstructor(), FloatConstructor(), BoolConstructor(), StrConstructor(),
        ListConstructor(), SetConstructor(), FrozenSetConstructor(),
        DictConstructor(), TupleConstructor(),
        void_function('Exception'), void_function("ValueError"),
        void_function('print'), int_function("abs"), int_function('len'),
        int_function('id'),
        int_function('ord'), num_function('pow'), num_function('sum'),
        bool_function('all'), bool_function('any'), bool_function('isinstance'),
        str_function('chr'), str_function('bin'), str_function('repr'), str_function('input'),
        FunctionType('open', returns=FileType),
        FunctionType('map', definition=map_definition),
        FunctionType('round', definition=round_definition),
        FunctionType('sorted', definition='identity'),
        FunctionType('reversed', definition='identity'),
        FunctionType('filter', definition='identity'),
        FunctionType('range', returns=lambda: ListType(False, IntType())),
        FunctionType('dir', returns=lambda: ListType(False, StrType())),
        FunctionType('max', returns='element'),
        FunctionType('min', returns='element'),
        FunctionType('enumerate', definition=enumerate_definition),
        FunctionType('zip', definition=zip_definition),
        FunctionType('__import__', returns=ModuleType),
        FunctionType('globals', returns=lambda: DictType([(StrType(), AnyType())])),
        FunctionType('classmethod', returns='identity'),
        FunctionType('staticmethod', returns='identity'),
    ]
})

BUILTIN_NAMES['__name__'] = StrType()


def get_builtin_module(name):
    """
    Given the name of the module, retrieve its TIFA representation.
    """
    return BUILTIN_MODULES.get(name)


def get_builtin_name(name):
    """
    Given a name, retrieve a copy of its built-in implementation.
    """
    if name in BUILTIN_NAMES:
        return BUILTIN_NAMES[name].clone_mutably()




"""
Main TIFA visitor-based algorithm here.

TODO: JoinedStr
"""

import ast
# TODO: FileType, DayType, TimeType,
from pedal.core.commands import system_error
from pedal.tifa.tifa_core import TifaCore, TifaAnalysis
from pedal.types.new_types import (Type, AnyType, ImpossibleType, FunctionType, GeneratorType,
                                   IntType, FloatType, BoolType, TupleType,
                                   ListType, StrType, SetType, DictType,
                                   ClassType, InstanceType, BuiltinConstructorType, NumType, NoneType,
                                   LiteralValue, LiteralInt, LiteralFloat, LiteralStr, LiteralBool,
                                   TypeUnion, widen_type, widest_type)
from pedal.types.new_types import is_subtype, specify_subtype
from pedal.types.normalize import get_pedal_type_from_value
from pedal.types.builtin import get_builtin_name
from pedal.types.operations import (VALID_UNARYOP_TYPES, apply_binary_operation, apply_unary_operation)
from pedal.tifa.constants import TOOL_NAME
from pedal.tifa.contexts import NewPath, NewScope
from pedal.tifa.feedbacks import (action_after_return, return_outside_function,
                                  write_out_of_scope, unconnected_blocks,
                                  iteration_problem, not_a_function,
                                  initialization_problem, unused_variable,
                                  overwritten_variable, iterating_over_non_list,
                                  iterating_over_empty_list, incompatible_types,
                                  parameter_type_mismatch, read_out_of_scope,
                                  type_changes, unnecessary_second_branch,
                                  recursive_call, multiple_return_types,
                                  possible_initialization_problem,
                                  incorrect_arity, else_on_loop_body,
                                  module_not_found, nested_function_definition,
                                  unused_returned_value, invalid_indexing,
                                  attribute_type_change)
from pedal.utilities.system import IS_AT_LEAST_PYTHON_38, IS_SKULPT


class Tifa(TifaCore, ast.NodeVisitor):
    """
    TIFA subclass for traversing an AST and finding common issues.
    You can instantiate this class, manipulate settings, and then process
    some code or AST.
    """

    def process_code(self, code, filename=None, reset=True):
        """
        Processes the AST of the given source code to generate a report.

        Args:
            code (str): The Python source code
            filename (str): The filename of the source code (defaults to
                the submissions' main file).
            reset (bool): Whether or not to reset the results from the
                previous analysis before running this one.
        Returns:
            Report: The successful or successful report object
        """
        if reset or self.analysis is None:
            self.analysis = TifaAnalysis()
        filename = filename or (self.report.submission.main_file if self.report.submission else 'student.py')
        self.line_offset = self.report.submission.line_offsets.get(filename, 0) if self.report.submission else 0

        # Attempt parsing - might fail!
        try:
            ast_tree = ast.parse(code, filename)
        except Exception as error:
            self.analysis.fail(error)
            system_error(TOOL_NAME, "Could not parse code: " + str(error),
                         report=self.report)
            return self.analysis
        # Attempt processing code - might fail!
        try:
            self.process_ast(ast_tree)
        except Exception as error:
            self.analysis.fail(error)
            system_error(TOOL_NAME, message="Successfully parsed but could not "
                                    "process AST: " + str(error),
                         report=self.report)
        # Return whatever we got
        return self.analysis

    def process_ast(self, ast_tree):
        """
        Given an AST, actually performs the type and flow analyses to return a
        report.

        Args:
            ast_tree (AST): The AST object
        """
        self.reset()
        # Traverse every node
        self.visit(ast_tree)

        # Update analysis, finish out the current scope.
        self.analysis.variables = self.name_map
        self._finish_scope()

        # Collect top level variables
        self._collect_top_level_variables()

        return self.analysis

    def visit(self, node):
        """
        Process this node by calling its appropriate visit_*

        Args:
            node (AST): The node to visit
        Returns:
            Type: The type calculated during the visit.
        """
        # Start processing the node
        self.node_chain.append(node)
        self.ast_id += 1

        # Actions after return?
        if len(self.scope_chain) > 1:
            return_state = self.find_variable_scope("*return")
            if return_state.exists and return_state.in_scope:
                if return_state.state.set == "yes":
                    self._issue(action_after_return(self.locate(),
                                                    report=self.report))

        # No? All good, let's enter the node
        self.final_node = node
        result = ast.NodeVisitor.visit(self, node)
        # If a node failed to return something, return the UNKNOWN TYPE
        if result is None:
            result = AnyType()
        self.analysis.node_types[node] = result

        # Pop the node out of the chain
        self.ast_id -= 1
        self.node_chain.pop()

        return result

    def _visit_nodes(self, nodes):
        """
        Visit all the nodes in the given list.

        Args:
            nodes (list): A list of values, of which any AST nodes will be
                          visited.
        """
        for node in nodes:
            if isinstance(node, ast.AST):
                self.visit(node)

    def _visit_collection_loop(self, node):
        was_empty = False
        # Handle the iteration list
        iter_list = node.iter
        iter_list_name = None
        if isinstance(iter_list, ast.Name):
            iter_list_name = iter_list.id
            if iter_list_name == "___":
                self._issue(unconnected_blocks(self.locate(iter_list), report=self.report))
            state = self.iterate_variable(iter_list_name, self.locate(iter_list))
            iter_type = state.type
        else:
            iter_type = self.visit(iter_list)

        if iter_type.is_empty:
            # TODO: It should check if its ONLY ever iterating over an empty list.
            # For now, only reports if we are NOT in a function
            was_empty = True
            if len(self.scope_chain) == 1:
                self._issue(iterating_over_empty_list(self.locate(iter_list), iter_list_name))

        iter_subtype = iter_type.iterate()
        if isinstance(iter_subtype, ImpossibleType):
            self._issue(iterating_over_non_list(self.locate(iter_list), iter_list_name, report=self.report))


        # Handle the iteration variable
        self.assign_target(node.target, iter_subtype, store_with_read=True)

        # Check that the iteration list and variable are distinct
        if isinstance(node.target, ast.Name) and node.target.id == iter_list_name:
            self._issue(iteration_problem(self.locate(node.target), node.target.id))

        return was_empty

    def break_starred(self, elements):
        leading, starred, trailing = [], [], []
        all_elements = iter(elements)
        for elt in all_elements:
            if isinstance(elt, ast.Starred):
                starred.append(elt)
                break
            else:
                leading.append(elt)
        else:
            return leading, starred, trailing
        for elt in all_elements:
            trailing.append(elt)
        return leading, starred, trailing

    # TODO: Properly handle assignments with subscripts
    def assign_target(self, target, target_type, operation=None, store_with_read=False):
        """
        Assign the type to the target, handling all kinds of assignment
        statements, including Names, Tuples/Lists, Subscripts, and
        Attributes.

        Args:
            target (ast.AST): The target AST Node.
            target_type (Type): The new TIFA type.
            operation (None | ast.op): The AugAssign operation, if there is one
            store_with_read: Whether to store the variable as a read variable, or just a write variable.

        Returns:

        """
        if isinstance(target, ast.Name):
            original_value_type = self.visit(target)
            if operation:
                new_target_type = apply_binary_operation(operation, original_value_type, target_type)
                if isinstance(new_target_type, ImpossibleType):
                    self._issue(incompatible_types(self.locate(), operation, original_value_type, target_type, report=self.report))
            else:
                new_target_type = target_type
            if store_with_read:
                self.store_read_variable(target.id, new_target_type)
            else:
                self.store_variable(target.id, new_target_type)
        elif isinstance(target, (ast.Tuple, ast.List)):
            original_type = self.visit(target)
            leading, starred, trailing = self.break_starred(target.elts)
            tt, ot = target_type.break_apart(), original_type.break_apart()
            for elt, elt_type, old_type in zip(leading, tt, ot):
                self.assign_target(elt, elt_type)
            if starred:
                # TODO: Handle starred node's type changes
                # if not is_subtype(target_type, old_type):
                    # self._issue(type_changes(self.locate(), 'an element of NODE', old_type.singular_name, elt_type.singular_name))
                self.assign_target(starred[0], target_type)
                for elt, elt_type, old_type in zip(trailing, tt, ot):
                    # BUG: Any trailing elements will be incorrectly offset, so won't work with finite length stuff
                    self.assign_target(elt, elt_type)
        elif isinstance(target, ast.Subscript):
            original_value_type = self.visit(target.value)
            origin = self.identify_caller(target.value)
            original_indexing_type = self.visit(target.slice)
            original_type = original_value_type.index(original_indexing_type)
            if operation:
                new_target_type = apply_binary_operation(operation, original_type, target_type)
                if isinstance(new_target_type, ImpossibleType):
                    self._issue(incompatible_types(self.locate(), operation, original_type, target_type, report=self.report))
            else:
                new_target_type = target_type
            original_type.set_index(original_indexing_type, new_target_type)
            if origin:
                origin_type = self.load_variable(origin)
                self.store_variable(origin, origin_type.type)
        elif isinstance(target, ast.Attribute):
            original_type = self.visit(target.value)
            origin = self.identify_caller(target.value)
            if operation:
                new_target_type = apply_binary_operation(operation, original_type, target_type)
                if isinstance(new_target_type, ImpossibleType):
                    self._issue(incompatible_types(self.locate(), operation, original_type, target_type, report=self.report))
            else:
                new_target_type = target_type
            assigned_type = original_type.add_attr(target.attr, new_target_type)
            if isinstance(assigned_type, ImpossibleType):
                self._issue(attribute_type_change(self.locate(), target.attr, original_type, new_target_type, report=self.report))
            if origin:
                origin_type = self.load_variable(origin)
                if origin_type:
                    self.store_variable(origin, origin_type.type)

    def visit_AnnAssign(self, node):
        """

        Args:
            node (ast.AnnAssign):

        Returns:

        """
        # Name, Attribute, or SubScript
        target = node.target
        # Type
        annotation = node.annotation
        annotation_type = self.evaluate_type(annotation)
        was_class_attribute = False
        # Optional assigned value
        value = node.value
        # 0 or 1, with 1 indicating pure names (not expressions)
        simple = node.simple

        # If it's a class attribute, then build up the type!
        if simple:
            current_scope = self.scope_chain[0]
            if current_scope in self.class_scopes:
                self.class_scopes[current_scope].add_attr(target.id, annotation)
                was_class_attribute = True
        # TODO: Allow setting for optional type coercion, or throw error
        if value:
            self.visit(value)
            # self.assign_target(target, self.visit(value))
        # Make local variable either way
        self.assign_target(target, annotation_type, store_with_read=was_class_attribute)

    def visit_Assign(self, node):
        """
        Simple assignment statement:
        __targets__ = __value__

        Args:
            node (ast.Assign): An Assign node
        Returns:
            None
        """
        # Handle value
        value_type = self.visit(node.value)
        # Apply value to all targets
        for target in node.targets:
            self.assign_target(target, value_type)

    def visit_AugAssign(self, node):
        """

        Args:
            node (ast.AugAssign):

        Returns:

        """
        # Handle value
        right = self.visit(node.value)
        if isinstance(node.target, ast.Name):
            self.load_variable(self.identify_caller(node.target))
        self.assign_target(node.target, right, node.op)

    def visit_Attribute(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        self.check_common_bad_lookups(value_type, node.attr, node.value)
        # Handle attr
        result = value_type.get_attr(node.attr)
        # Set up self if this was a function (because now it's a method!)
        # TODO: Handle static/class functions differently I guess?
        if isinstance(result, FunctionType):
            result = result.as_method(value_type)
        return result

    def visit_BinOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        right = self.visit(node.right)
        operation = node.op

        new_target_type = apply_binary_operation(operation, left, right)
        if isinstance(new_target_type, ImpossibleType):
            self._issue(incompatible_types(self.locate(), operation, left, right, report=self.report))
        return new_target_type

    def visit_Bool(self, node):
        """
        Visit a constant boolean value.

        Args:
            node (ast.AST): The boolean value Node.

        Returns:
            Type: A Bool type.
        """
        return BoolType()

    def visit_BoolOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        values = [self.visit(value) for value in node.values]
        # Handle operation
        # Actually, the operations are always the same behavior!
        # Handle truthiness
        if self.get_setting('truthiness_returns_booleans'):
            return BoolType()
        elif not values:
            return ImpossibleType()
        # All same type? Then return the first one!
        elif all(is_subtype(values[0], other) for other in values[1:]):
            return values[0]
        # Oh no, type union is necessary!
        else:
            return TypeUnion(values)

    def load_root_variable(self, node, position=None):
        root_name = self.identify_caller(node)
        if root_name is not None:
            variable = self.find_variable_scope(root_name)
            if variable.exists:
                return variable.state
        return None

    def visit_Call(self, node):
        # Handle func part
        function_type = self.visit(node.func)
        # TODO: Need to grab the actual type in some situations
        original_callee = self.load_root_variable(node)

        # Handle args
        arguments = [self.visit(arg) for arg in node.args] if node.args else []
        keywords = [(kwarg.arg, self.visit(kwarg.value))
                    for kwarg in node.keywords] if node.keywords else {}

        # Check special common mistakes

        if isinstance(function_type, BuiltinConstructorType):
            constructor = function_type.definition
            return constructor(self, function_type, original_callee,
                               arguments, keywords, self.locate())
        if isinstance(function_type, FunctionType):
            # Test if we have called this definition before
            if function_type.definition not in self.definition_chain:
                self.definition_chain.append(function_type.definition)
                # Function invocation
                result = function_type.definition(self, function_type, original_callee,
                                                  arguments, keywords, self.locate())
                self.definition_chain.pop()
                return result
            else:
                self._issue(recursive_call(self.locate(), original_callee, report=self.report))
        elif isinstance(function_type, ClassType):
            constructor = function_type.get_constructor().definition
            self.definition_chain.append(constructor)
            new_instance = constructor(self, constructor, original_callee, arguments, keywords, self.locate())
            self.definition_chain.pop()
            if '__init__' in function_type.fields:
                initializer = function_type.fields['__init__'].as_method(new_instance)
                if isinstance(initializer, FunctionType):
                    self.definition_chain.append(initializer)
                    initializer.definition(self, initializer, new_instance, [new_instance] + arguments, keywords, self.locate())
                    self.definition_chain.pop()
            return new_instance
        elif isinstance(function_type, (IntType, FloatType, NumType, ListType,
                                        DictType, SetType, TupleType, InstanceType,
                                        StrType, BoolType, NoneType, LiteralValue)):
            if original_callee is None:
                self._issue(not_a_function(self.locate(), "a value", function_type, report=self.report))
            else:
                self._issue(not_a_function(self.locate(), original_callee.name, function_type, report=self.report))
            return ImpossibleType()
        return AnyType()

    def visit_ClassDef(self, node):
        """

        Args:
            node:
        """
        class_name = node.name
        parents = [self.visit(base) for base in node.bases]
        new_class_type = ClassType(class_name, {}, parents)
        self.store_variable(class_name, new_class_type)
        definitions_scope = self.scope_chain[:]
        class_scope = NewScope(self, definitions_scope, class_type=new_class_type)
        # TODO: Handle metaclass... somehow?
        #self._visit_nodes(node.keywords)
        with class_scope:
            self._visit_nodes(node.body)
        new_class_type = self.apply_decorators(class_name, new_class_type, node.decorator_list)
        return new_class_type

    def apply_decorators(self, new_name, new_type, decorators):
        for decorator in reversed(decorators):
            old_type = self.load_variable(new_name)
            decorator_type = self.visit(decorator)
            original_callee = self.load_root_variable(decorator)
            new_type = decorator_type.definition(self, decorator_type, original_callee,
                                                 [new_type], {}, self.locate())
            self.store_variable(new_name, new_type)
        return new_type

    def visit_Compare(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        comparators = [self.visit(compare) for compare in node.comparators]

        # Handle ops
        for op, right in zip(node.ops, comparators):
            if isinstance(op, (ast.Eq, ast.NotEq, ast.Is, ast.IsNot)):
                continue
            elif isinstance(op, (ast.Lt, ast.LtE, ast.GtE, ast.Gt)):
                if type(right) in left.orderable:
                    continue
            elif isinstance(op, (ast.In, ast.NotIn)):
                if right.allows_membership(left):
                    continue
            self._issue(incompatible_types(self.locate(), op, left, right, report=self.report))
        return BoolType()

    def visit_comprehension(self, node):
        """

        Args:
            node:
        """
        self.fill_in_location(node, self.node_chain[-2])
        self._visit_collection_loop(node)
        # Handle ifs, unless they're blank (None in Skulpt :)
        if node.ifs:
            self.visit_statements(node.ifs)

    def fill_in_location(self, node, source_node):
        node.lineno = source_node.lineno
        node.col_offset = source_node.col_offset
        if IS_AT_LEAST_PYTHON_38:
            node.end_lineno = source_node.end_lineno
            node.end_col_offset = source_node.end_col_offset
        else:
            node.end_lineno = source_node.lineno
            node.end_col_offset = source_node.col_offset

    def visit_Dict(self, node):
        """
        Three types of dictionaries
        - empty
        - uniform type
        - record
        TODO: Handle records appropriately
        """
        items = [(self.visit(key), self.visit(value))
                 for key, value in zip(node.keys, node.values) if key is not None]
        # Unpack starred dictionaries
        for key, value in zip(node.keys, node.values):
            if key is None:
                value_type = self.visit(value)
                if isinstance(value_type, DictType):
                    items.extend([
                        (k, v) for k, v in value_type.element_types
                    ])
        # Empty dictionary
        if not items:
            return DictType([])
        # All literal keys
        if all(isinstance(k, LiteralValue) for k, _ in items):
            return DictType([(k, v) for k, v in items])
        # Find if all matched type
        first_key = widest_type([key for key, value in items])
        first_value = widest_type([value for key, value in items])
        if first_key is not None and first_value is not None:
            return DictType([(first_key, first_value)])
        else:
            # Resort to just matching the types?
            return DictType([(k, v) for k, v in items])

    def visit_DictComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        generator_scope = NewScope(self, self.scope_chain[:])
        with generator_scope:
            self._visit_nodes(node.generators)
            keys = self.visit(node.key)
            values = self.visit(node.value)
            return DictType([(keys, values)])

    def visit_Expr(self, node):
        """
        Any expression being used as a statement.

        Args:
            node (AST): An Expr node

        Returns:

        """
        value = self.visit(node.value)
        if isinstance(node.value, ast.Call) and not isinstance(value, NoneType):
            # TODO: Helper function to get name with title ("append method")
            if isinstance(node.value.func, ast.Name):
                call_type = 'function'
            else:
                call_type = 'method'
            name = self.identify_caller(node.value)
            self._issue(unused_returned_value(self.locate(), name,
                                              call_type, value))

    def visit_For(self, node):
        """

        Args:
            node:
        """
        was_empty = self._visit_collection_loop(node)
        # Handle the bodies
        # if not was_empty:
        # this_path_id = self.path_chain[0]
        # non_empty_path = NewPath(self, this_path_id, "f")
        # with non_empty_path:
        self.visit_statements(node.body)
        self.visit_statements(node.orelse)

    def visit_FunctionDef(self, node):
        """

        Args:
            node:

        Returns:

        """
        function_name = node.name
        position = self.locate()
        definitions_scope = self.scope_chain[:]
        definition = self.make_function(function_name, definitions_scope, position, node)
        function = FunctionType(function_name, definition=definition)
        self.store_variable(function_name, function)

        if len(self.node_chain) > 2:
            self._issue(nested_function_definition(self.locate(), function_name))

        return self.apply_decorators(function_name, function, node.decorator_list)

    def make_function(self, function_name, definitions_scope, position, node):
        is_lambda = function_name is None

        class SkipType:
            pass
        # Pull apart args into a coherent set of data PRIOR to execution
        args = node.args
        pos_parameters = args.posonlyargs + args.args if IS_AT_LEAST_PYTHON_38 and not IS_SKULPT else args.args
        none_padding = [SkipType] * (len(pos_parameters) - len(args.defaults))
        pos_defaults = none_padding + [self.visit(d) for d in args.defaults]
        kwarg_parameter = args.kwarg
        vararg_parameter = args.vararg

        # TODO: Finish kwargs
        named_parameters = args.kwonlyargs
        named_defaults = [self.visit(d) if d is not None else AnyType() for d in args.kw_defaults]
        expected_returns = self.evaluate_type(node.returns) if not is_lambda and node.returns else None

        def definition(tifa, function, callee_name, arguments, named_arguments, call_position):
            function_scope = NewScope(self, definitions_scope)
            with function_scope:
                # Process arguments
                if len(arguments) + len(named_arguments) != len(pos_parameters) and not vararg_parameter:
                    self._issue(incorrect_arity(self.locate(), function_name,
                                                len(pos_parameters), len(arguments), report=self.report))
                for parameter, default, argument in zip(pos_parameters, pos_defaults, arguments):
                    parameter_name = parameter.arg
                    if parameter.annotation:
                        annotation = self.evaluate_type(parameter.annotation)
                        if is_subtype(argument, annotation):
                            specify_subtype(annotation, argument)
                        else:
                            self._issue(parameter_type_mismatch(self.locate(), parameter_name,
                                                                annotation, argument))
                    elif default is not SkipType:
                        if is_subtype(argument, default):
                            specify_subtype(default, argument)
                        else:
                            self._issue(parameter_type_mismatch(self.locate(), parameter_name,
                                                                default, argument))
                    if argument is not None:
                        argument_type = argument.clone_mutably()
                        self.create_variable(parameter_name, argument_type, position)
                # Too many arguments
                if len(pos_parameters) < len(arguments):
                    if vararg_parameter:
                        the_rest = arguments[len(pos_parameters):]
                        self.create_variable(vararg_parameter.arg, TupleType(the_rest), position)
                # Not enough arguments
                if len(pos_parameters) > len(arguments):
                    for parameter in pos_parameters[len(arguments):]:
                        if parameter.annotation:
                            annotation = self.evaluate_type(parameter.annotation)
                        else:
                            annotation = AnyType()
                        self.create_variable(parameter.arg, annotation, position)
                if is_lambda:
                    return self.visit(node.body)

                self.visit_statements(node.body)
                return_state = self.find_variable_scope("*return")
                return_value = NoneType()
                # TODO: Figure out if we are not returning something when we should
                # If the pseudo variable exists, we load it and get its type
                if return_state.exists and return_state.in_scope:
                    return_state = self.load_variable("*return", call_position)
                    return_value = return_state.type
                    if expected_returns:
                        if not is_subtype(return_value, expected_returns):
                            self._issue(multiple_return_types(return_state.position,
                                                              expected_returns.singular_name,
                                                              return_value.singular_name,
                                                              report=self.report))
            return return_value
        return definition

    def evaluate_type(self, node):
        string_literal = False
        if isinstance(node, ast.Str):
            string_literal = node.s
        elif isinstance(node, ast.Constant) and isinstance(node.value, str):
            string_literal = node.value
        if string_literal is not False:
            if self.get_setting('evaluate_string_literal_types'):
                # TODO: Handle string literal types properly!
                pass
            else:
                return LiteralStr(string_literal)
        elif isinstance(node, ast.List):
            if node.elts:
                return ListType(False, self.evaluate_type(node.elts[0]))
            else:
                return ListType(True)
        elif isinstance(node, ast.Set) and node.elts:
            return TypeUnion([self.evaluate_type(e) for e in node.elts])
        elif isinstance(node, ast.Tuple) and node.elts:
            return TupleType([self.evaluate_type(e) for e in node.elts])
        elif isinstance(node, ast.Dict):
            if node.keys and node.values:
                return DictType([(self.evaluate_type(k), self.evaluate_type(v))
                                 for k, v in zip(node.keys, node.values)])
            else:
                return DictType([])
        else:
            evaluated = self.visit(node)
            return evaluated.as_type(self, self.locate(node))

    def visit_GeneratorExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        generator_scope = NewScope(self, self.scope_chain[:])
        with generator_scope:
            self._visit_nodes(node.generators)
            return GeneratorType(False, self.visit(node.elt))

    def visit_If(self, node):
        """

        Args:
            node:
        """
        # Visit the conditional
        self.visit(node.test)

        if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.Pass):
            self._issue(unnecessary_second_branch(self.locate()))
        elif len(node.body) == 1 and isinstance(node.body[0], ast.Pass):
            if node.orelse:
                self._issue(unnecessary_second_branch(self.locate()))

        # Visit the bodies
        this_path_id = self.path_chain[0]
        if_path = NewPath(self, this_path_id, "i")
        with if_path:
            for statement in node.body:
                self.visit(statement)
        else_path = NewPath(self, this_path_id, "e")
        with else_path:
            for statement in node.orelse:
                self.visit(statement)

        # TODO: Unconditional return

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, if_path.id, else_path.id)

    def visit_IfExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Visit the conditional
        self.visit(node.test)

        # Visit the body
        body = self.visit(node.body)

        # Visit the orelse
        orelse = self.visit(node.orelse)

        if is_subtype(body, orelse):
            return body

        # TODO: Union type?
        return TypeUnion([body, orelse])

    def visit_Import(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            asname = alias.asname or alias.name
            module_type = self.load_module(alias.name)
            self.store_variable(asname, module_type)

    def visit_ImportFrom(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            if alias.name == '*':
                module_type = self.load_module(node.module)
                for field, value in module_type.fields.items():
                    self.store_read_variable(field, value)
            else:
                if node.module is None:
                    asname = alias.asname or alias.name
                    module_type = self.load_module(alias.name)
                else:
                    module_name = node.module
                    asname = alias.asname or alias.name
                    module_type = self.load_module(module_name)
                name_type = module_type.get_attr(alias.name)
                self.store_variable(asname, name_type)

    def visit_Lambda(self, node):
        """

        Args:
            node:

        Returns:

        """
        position = self.locate()
        definitions_scope = self.scope_chain[:]
        definition = self.make_function(None, definitions_scope, position, node)
        return FunctionType("*lambda", definition=definition)

    def visit_List(self, node):
        """

        Args:
            node:

        Returns:

        """
        if not node.elts:
            return ListType(True)

        # All literal keys
        elements = [self.visit(v) for v in node.elts]
        if all(isinstance(v, LiteralValue) for v in elements):
            # TODO: Allow type union instead?
            return ListType(False, elements[0].promote())
        # Find if all matched type
        first_value = widest_type(elements)
        if first_value is not None:
            return ListType(False, first_value)
        else:
            # Resort to just matching the types?
            return ListType(False, TypeUnion(elements))

    def visit_ListComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        generator_scope = NewScope(self, self.scope_chain[:])
        with generator_scope:
            self._visit_nodes(node.generators)
            return ListType(False, self.visit(node.elt))

    def visit_NameConstant(self, node):
        """

        Args:
            node:

        Returns:

        """
        value = node.value
        if isinstance(value, bool):
            return LiteralBool(value)
        else:
            return NoneType()

    def visit_Name(self, node):
        """

        Args:
            node:

        Returns:

        """
        name = node.id
        if name == "___":
            self._issue(unconnected_blocks(self.locate()))
        if isinstance(node.ctx, ast.Load):
            if name == "True" or name == "False":
                return LiteralBool(name == "True")
            elif name == "None":
                return NoneType()
            else:
                variable = self.find_variable_scope(name)
                builtin = get_builtin_name(name)
                if not variable.exists and builtin:
                    return builtin
                else:
                    state = self.load_variable(name)
                    return state.type
        else:
            variable = self.find_variable_scope(name)
            if variable.exists:
                return variable.state.type
            else:
                return AnyType()

    def visit_Num(self, node):
        """

        Args:
            node:

        Returns:

        """
        return LiteralInt(node.n) if isinstance(node.n, int) else LiteralFloat(node.n)

    def visit_Constant(self, node) -> Type:
        """ Handle new 3.8's Constant node """
        return get_pedal_type_from_value(node.value, self.evaluate_type)

    def visit_Return(self, node):
        """

        Args:
            node:
        """
        if len(self.scope_chain) == 1:
            self._issue(return_outside_function(self.locate(), report=self.report))
        # TODO: Unconditional return inside loop
        if node.value is not None:
            self.return_variable(self.visit(node.value))
        else:
            self.return_variable(NoneType())

    def visit_Set(self, node):
        # Fun fact, it's impossible to make a literal empty set
        if not node.elts:
            return SetType(True)

        # All literal keys
        elements = [self.visit(v) for v in node.elts]
        if all(isinstance(v, LiteralValue) for v in elements):
            # TODO: Allow type union instead?
            return SetType(False, elements[0].promote())
        # Find if all matched type
        first_value = widest_type(elements)
        if first_value is not None:
            return SetType(False, first_value)
        else:
            # Resort to just matching the types?
            return SetType(False, TypeUnion(elements))

    def visit_SetComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        generator_scope = NewScope(self, self.scope_chain[:])
        with generator_scope:
            self._visit_nodes(node.generators)
            return SetType(False, self.visit(node.elt))

    def visit_statements(self, nodes):
        """

        Args:
            nodes:

        Returns:

        """
        # TODO: Check for pass in the middle of a series of statement
        if any(isinstance(node, ast.Pass) for node in nodes):
            pass
        return [self.visit(statement) for statement in nodes]

    def visit_Str(self, node):
        """

        Args:
            node:

        Returns:

        """
        return LiteralStr(node.s)

    def visit_JoinedStr(self, node):
        values = [self.visit(expr) for expr in node.values]
        # The result will be all StrType
        return StrType(is_empty=all(isinstance(n, (StrType, LiteralStr)) and n.is_empty
                                    for n in values))

    def visit_FormattedValue(self, node):
        value = self.visit(node.value)
        if isinstance(value, StrType):
            return value
        else:
            return StrType(is_empty=False)

    def visit_Subscript(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        slice_type = self.visit(node.slice)
        result = value_type.index(slice_type)
        if isinstance(result, ImpossibleType):
            self._issue(invalid_indexing(self.locate(), value_type, slice_type))
        if isinstance(node.slice, ast.Slice):
            return value_type.shallow_clone()
        else:
            return result

    def visit_Slice(self, node):
        """ Handles a slice by visiting its components; cannot return a value
        because the slice is always the same type as its value, which is
        not available on the Slice node itself. """
        if node.lower is not None:
            self.visit(node.lower)
        if node.upper is not None:
            self.visit(node.upper)
        if node.step is not None:
            self.visit(node.step)

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_ExtSlice(self, node):
        return TupleType([self.visit(d) for d in node.dims])

    def visit_Starred(self, node):
        return self.visit(node.value).break_apart()

    def visit_Tuple(self, node) -> TupleType:
        # Fun fact, it's impossible to make a literal empty set
        if not node.elts:
            return TupleType(True)

        # All literal keys
        return TupleType([self.visit(v) for v in node.elts])

    def visit_UnaryOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle operand
        operand_type = self.visit(node.operand)
        return apply_unary_operation(node.op, operand_type)

    def visit_While(self, node):
        """

        Args:
            node:
        """
        # Visit conditional
        self.visit(node.test)

        # Visit the bodies
        this_path_id = self.path_id
        # One path is that we never enter the body
        empty_path = NewPath(self, this_path_id, "e")
        with empty_path:
            pass
        # Another path is that we loop through the body and check the test again
        body_path = NewPath(self, this_path_id, "w")
        with body_path:
            for statement in node.body:
                self.visit(statement)
            # Revisit conditional
            self.visit(node.test)
        # If there's else bodies (WEIRD) then we should check them afterwards
        if node.orelse:
            self._issue(else_on_loop_body(self.locate()))
            for statement in node.orelse:
                self.visit(statement)

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, body_path.id, empty_path.id)

    def visit_With(self, node):
        """

        Args:
            node:
        """
        for item in node.items:
            type_value = self.visit(item.context_expr)
            if item.optional_vars is not None:
                self.assign_target(item.optional_vars, type_value)
        # Handle the bodies
        self.visit_statements(node.body)



"""
Utilities for interfacing with ProgSnap2 data files, both in SQL format and
in Zipped-CSV format.

Filters can be: Exact string match, regex, or function
    TODO: So far only really support exact string match

    TODO: No support for link_context

progsnap = SqlProgSnap2("test/datafiles/progsnap2_3.db")
fun_student_edits = progsnap.get_events(
    event_filter={'EventType': 'File.Edit'},
    link_filters={
        'Subject': {
            'X-IsStaff': "False",
        },
        'Assignment': {
            'X-Name': "Fun%"
        },
    },
    link_selections={
        'Subject': {
            'X-Email': 'student_email',
            'X-Name.First': 'student_first'
            'X-Name.Last': 'student_last',
        },
        'Assignment': {
            'X-Name': 'assignment_name',
            'X-URL': 'assignment_url',
            'X-Code.OnRun': 'on_run'
        }
    }
)
"""
import hashlib
import os
import pickle
import sqlite3
import re


class BaseProgSnap2:
    def __init__(self, path: str):
        self.path = path

    def get_events(self, event_filter=None, link_filters=None,
                   link_selections=None, with_code=True):
        raise NotImplemented("Cannot execute get_events on base ProgSnap2 Format")

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class SqlProgSnap2(BaseProgSnap2):
    PROFILES = {
        'blockpy': dict(
            link_filters={
                'Subject': {
                    'X-IsStaff': "False",
                },
                "Assignment": {
                    #"X-Name": "Midterm 2.4"
                #    "X-Course.Id": "4"
                }
            },
            link_selections={
                'Subject': {
                    #'X-Pokemon': 'student_pokemon',
                    'X-Email': 'student_email',
                    'X-Name.First': 'student_first',
                    'X-Name.Last': 'student_last',
                },
                'Assignment': {
                    'X-Name': 'assignment_name',
                    'X-URL': 'assignment_url',
                    'X-Code.OnRun': 'on_run'
                }
            },
            link_primary={
                'user': 'student_email'
            },
        ),
        'blockpy_consenting': dict(
            link_filters={
                # '': [
                #     ("LinkAssignment.`X-URL` LIKE ?", "%read%"),
                #     ("LinkAssignment.`X-URL` LIKE ?", "%quiz%"),
                # ]
            },
            link_selections={
                'Assignment': {
                    'X-Name': 'assignment_name',
                    'X-URL': 'assignment_url',
                    'X-Code.OnRun': 'on_run'
                }
            },
            link_primary={
                'user': 'SubjectID'
            },
        ),
    }

    def set_profile(self, profile):
        if profile not in self.PROFILES:
            raise ValueError(f"Unknown ProgSnap Profile specified: {profile}")
        self.profile = profile

    def __init__(self, path: str, cache=False, status_update=None):
        super().__init__(path)
        self._connection = sqlite3.connect(self.path)
        self._cursor = self._connection.cursor()
        self.profile = None
        self.cache = cache
        # To see executed queries:
        # self._connection.set_trace_callback(print)
        if status_update:
            self._connection.set_progress_handler(status_update, 1000)

    def close(self):
        self._connection.close()

    def _merge(self, key, overrides):
        result = {}
        for k, v in self.PROFILES.get(self.profile, {}).get(key, {}).items():
            result[k] = v
        if overrides is not None:
            for k, v in overrides.items():
                result[k] = v
        return result

    def get_code(self, query):
        #flat_filters = {k + k2: v
        #                for k, vs in link_filters.items()
        #                for k2, v in vs
        #                }
        #codes = tuple(sorted(event_filter.items() | flat_filters))
        encoded = "pedal_cache_"+hashlib.md5(query.encode('utf-8')).hexdigest()+".pickle"
        return encoded

    def get_events(self, event_filter=None, link_filters=None,
                   link_selections=None, with_code=True, limit=None):
        # Load in profile defaults
        link_selections = self._merge('link_selections', link_selections)
        link_filters = self._merge('link_filters', link_filters)
        # Setup data
        tables = ['MainTable']
        filters = []  # "MainTable.EventType=?"
        selections = ['MainTable.EventID', 'MainTable.ClientTimestamp']
        fields = ['event_id', 'client_timestamp']
        if 'Subject' not in link_selections:
            selections.append('MainTable.SubjectID')
            fields.append('subject_id')
        data = []
        # Add in event filters
        for column, value_filter in event_filter.items():
            if isinstance(value_filter, str):
                filters.append(f"MainTable.`{column}`=?")
                data.append(value_filter)
        # Add in code
        if with_code:
            tables.append("CodeState")
            filters.append(f"MainTable.CodeStateID=CodeState.ID")
            selections.append("CodeState.Contents as submission_code")
            fields.append("submission_code")
        # Add in all needed link tables
        for table in (link_filters.keys() | link_selections.keys()):
            if table:
                tables.append('Link' + table)
                filters.append(f"MainTable.`{table}ID`=Link{table}.`{table}Id`")
        # Add in link table filters
        for table, table_filters in link_filters.items():
            # Blank string is generic filters
            if table == "":
                if isinstance(table_filters, str):
                    table_filters = [table_filters]
                for value_filter, v in table_filters:
                    filters.append(value_filter)
                    data.append(v)
            else:
                # Otherwise, assume link table -> column filter
                for column, value_filter in table_filters.items():
                    if isinstance(value_filter, str):
                        # TODO: Make this more robust
                        if "%" in value_filter:
                            filters.append(f"Link{table}.`{column}` LIKE ?")
                        else:
                            filters.append(f"Link{table}.`{column}`=?")
                        data.append(value_filter)
        # Add in Contextual tables
        for table, table_selections in link_selections.items():
            for column, alias in table_selections.items():
                selections.append(f"Link{table}.`{column}` AS {alias}")
                fields.append(alias)
        # Prepare actual filter query string
        if filters:
            filters = "WHERE " + " AND ".join(filters)
        else:
            filters = ""
        # And a limit for testing purposes
        query_limit = ""
        if limit:
            query_limit = f"LIMIT {limit}"
        # And execute
        query = f"""
            SELECT {', '.join(selections)}
            FROM {', '.join(tables)}
            {filters}
            {query_limit}
        """
        # Is it is in our cache?
        if isinstance(self.cache, str):
            key = self.get_code(query)
            cache_path = os.path.join(self.cache, key)
            if os.path.exists(cache_path):
                with open(cache_path, 'rb') as cache_file:
                    return pickle.load(cache_file)
        result = self._cursor.execute(query, data)
        result = [dict(zip(fields, row)) for row in result.fetchall()]
        # Store it in our cache for the future
        if isinstance(self.cache, str):
            key = self.get_code(query)
            cache_path = os.path.join(self.cache, key)
            with open(cache_path, 'wb') as cache_file:
                pickle.dump(result, cache_file)
        # Process result
        return result


class ZipProgSnap2(BaseProgSnap2):
    pass



"""
Generic runtime exception feedback class.

TODO: Use the FeedbackRegistry to properly dispatch the appropriate runtime_error type, instead of reusing the base object
"""
from pedal.core.formatting import wrap_fields
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.sandbox.data import format_contexts
from pedal.utilities.exceptions import KeyError, get_exception_name
from pedal.core.feedback import FeedbackResponse
from pedal.sandbox import TOOL_NAME
from pedal.utilities.text import add_indefinite_article


class runtime_error(FeedbackResponse):
    """
    Used to create all runtime errors.

    Attributes:
        exception (Exception): The original exception.
        exception_name (str): The original name of the exception.
        exception_message (str): the original error message from the exception.
        traceback (:py:class:`~pedal.utilities.exceptions.ExpandedTraceback`):
            An enhanced version of the builtin Traceback object. Has a number
            of additional fields.
        traceback_message (str): A nicely formatted version of the traceback.
        context (list[:py:class:`~pedal.sandbox.data.SandboxContext`]): The
            history of inputs, outputs, executed code, and other information
            from when this runtime error occurred. Usually just a single
            element, but sometimes longer if created in a grouping context.
            If the code was run from a file, then the filename is given.

    """
    tool = TOOL_NAME
    category = FeedbackResponse.CATEGORIES.RUNTIME
    kind = FeedbackResponse.KINDS.MISTAKE
    justification = "A runtime error occurred during execution of some code."
    version = '1.1.0'
    constant_fields = {"suggestion": ""}
    message_template = (
        "{context_message}\n"
        "{exception_name_proper} occurred:\n\n"
        "{exception_message}\n\n"
        "{traceback_preamble}{traceback_message}\n"
        "{suggestion_message}"
    )

    def __init__(self, exception, context, traceback, location, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        exception_name = get_exception_name(exception)
        exception_name_proper = add_indefinite_article(exception_name)
        exception_message = str(exception)
        exception_message = exception_message[0].upper() + exception_message[1:] if exception_message else ""
        if type(exception) not in EXCEPTION_FF_MAP:
            title = exception_name
        else:
            title = EXCEPTION_FF_MAP[type(exception)].title
        location = Location(location)
        traceback_stack = traceback.build_traceback()
        traceback_message = traceback.format_traceback(traceback_stack, report.format) or ""
        traceback_preamble = f"The traceback was:\n" if traceback_message else ""
        context_message = format_contexts(context, report.format)
        fields = {'exception': exception,
                  'exception_name': exception_name,
                  'exception_name_proper': exception_name_proper,
                  'exception_message': report.format.exception(exception_message),
                  'location': location,
                  'traceback': traceback,
                  'traceback_preamble': traceback_preamble,
                  'traceback_stack': traceback_stack,
                  'traceback_message': traceback_message,
                  'context': context,
                  'context_message': context_message}
        fields['suggestion_message'] = self.constant_fields['suggestion'].format(**wrap_fields(report.format, fields))
        super().__init__(fields=fields, title=title,
                         location=location, **kwargs)


class type_error(runtime_error):
    """ Type Error """
    title = "Type Error"
    constant_fields = {'suggestion': "Type errors occur when you use an operator or "
                        "function on the wrong type of value. For example, "
                        "using `+` to add to a list (instead of `.append`), or "
                        "dividing a string by a number.\n\n"
                        "Suggestion: To fix a type error, you should trace "
                        "through your code. Make sure each expression has the "
                        "type you expect it to have."}


class name_error(runtime_error):
    """ Runtime NameError """
    title = "Name Error"
    constant_fields = {'suggestion': "A name error means you have used a variable that has "
                        "no value.  You may have a typo, so check the "
                        "spelling. Or, you may have forgotten to initialize a "
                        "variable.\n\n"
                        "Suggestion: Trace your code and make sure each "
                        "variable was set before it was read."}


class value_error(runtime_error):
    """ Runtime ValueError """
    title = "Value Error"
    constant_fields = {'suggestion': "A ValueError occurs when you pass the wrong type of "
                        "value to a function. For example, you try to convert "
                        "a string without numbers to an integer (like "
                        "`int('Five')`).\n\n"
                        "Suggestion: Read the error message to see which "
                        "function had the issue. Check what type the function "
                        "expects. Then trace your code to make sure you pass "
                        "in that type."}


class attribute_error(runtime_error):
    """ Runtime AttributeError """
    title = "Attribute Error"
    constant_fields = {'suggestion': "An AttributeError means you used an attribute or "
                        "method that does not exist. For example, you wrote "
                        "`text.delete()` even though there is no `delete` "
                        "method.\n\n"
                        "Suggestion: Make sure that you spelled the method or "
                        "attribute correctly. Then, trace your code to check "
                        "that the dot's left expression is the correct type."}


class index_error(runtime_error):
    """ Runtime IndexError """
    title = "Index Error"
    constant_fields = {'suggestion': "An IndexError means that you indexed past the end of "
                        "a string or a list.  For example, if you access index "
                        "5 in a list with 3 items.\n\n"
                        "Suggestion: Remember that the first position in a "
                        "list or string is 0. Often, you will be off by just "
                        "one index position, so check your math. Also, make "
                        "sure the list or string has the right value."}


class zero_division_error(runtime_error):
    """ Runtime ZeroDivisionError """
    title = "Division by Zero Error"
    constant_fields = {'suggestion': "A ZeroDivisionError means you divided by 0. The "
                        "denominator (the right side of a division expression) "
                        "cannot be 0.\n\n"
                        "Suggestion: Check that you are dividing by the "
                        "correct value. Or, you might need to add an `if` "
                        "statement to handle the zero case."}


class indentation_error(runtime_error):
    """ Syntax IndentationError """
    title = "Indentation Error"
    constant_fields = {'suggestion': "An IndentationError means you have not indented your "
                        "code correctly. You have too many or too few spaces.\n\n"
                        "Suggestion: Check the line number, and the lines "
                        "before and after. Check the body of each function "
                        "definition, `if` statement, and `for` loop. Remember, "
                        "all the statements INSIDE of a body have the same "
                        "indentation."}


class import_error(runtime_error):
    """ Runtime ImportError """
    title = "Import Error"
    constant_fields = {'suggestion': "An ImportError means you tried to import a module "
                        "that does not exist. You might have a typo or a file "
                        "might be in the wrong location.\n\n"
                        "Suggestion: Check the spelling and capitalization of "
                        "the module's name. If you are importing a file, then "
                        "check that it is in the correct folder."}


class io_error(runtime_error):
    """ Runtime IOError """
    title = "Input/Output Error"
    constant_fields = {'suggestion': "An IOError means you tried to open a file that was "
                        "not available.\n\n"
                        "Suggestion: Make sure that the file is in the "
                        "correct folder. Then, make sure you spelled the "
                        "filename correctly."}


class key_error(runtime_error):
    """ Runtime KeyError """
    title = "Key Error"
    constant_fields = {'suggestion': "A KeyError means you accessed a non-existent key in "
                        "a dictionary.\n\n"
                        "Suggestion: First, check that you spelled the key "
                        "correctly. Make sure the key has the right type and "
                        "value. Then, check that the dictionary actually has "
                        "that key."}


class memory_error(runtime_error):
    """ Runtime MemoryError """
    title = "Memory Error"
    constant_fields = {'suggestion': "A MemoryError means your program ran out of mental "
                        "space.\n\n"
                        "Suggestion: You might have an infinite loop. Or, you "
                        "might not be filtering your data enough."}


class timeout_error(runtime_error):
    """ Runtime TimeoutError """
    title = "Timeout Error"
    constant_fields = {'suggestion': "A TimeoutError means your program took too long to "
                        "run.\n\n"
                        "Suggestion: Check that you do not have an infinite "
                        "loop."}


EXCEPTION_FF_MAP = {
    TypeError: type_error,
    NameError: name_error,
    ValueError: value_error,
    AttributeError: attribute_error,
    IndexError: index_error,
    ZeroDivisionError: zero_division_error,
    IndentationError: indentation_error,
    ImportError: import_error,
    IOError: io_error,
    KeyError: key_error,
    MemoryError: memory_error,
    TimeoutError: timeout_error,
}



"""
Python Type Inferencer and Flow Analyzer (TIFA)

TIFA uses a number of simplifications of the Python language.
  * Variables cannot change type
  * Variables cannot be deleted
  * Complex types have to be homogenous
  * No introspection or reflective characteristics
  * No dunder methods
  * No closures (maybe?)
  * You cannot write a variable out of scope
  * You cannot read a mutable variable out of scope
  * No multiple inheritance

Additionally, it reads the following as issues:
  * Cannot read a variable without having first written to it.
  * Cannot rewrite a variable unless it has been read.

Important concepts:

.. glossary::

    Issue
        A problematic situation in the submitted code that will be reported
        but may not stop the execution. However, when an Issue occurs,
        any results may be invalid.

    Error
        A situation in execution that terminates the program.

    Name
        A name of a variable

    Scope
        The context of a function, with its own namespaces. Represented
        internally using numeric IDs (Scope IDs).

    Scope Chain
        A stack of scopes, with the innermost scope on top.

    Fully Qualified Name
        A string representation of a variable and its scope
        chain, written using "/". For example: 0/1/4/my_variable_name

    Path
        A single path of execution through the control flow; every program
        has at least one sequential path, but IFs, FORs, WHILEs, etc. can
        cause multiple paths. Paths are represented using numeric IDs (Path
        IDs).

    State
        Information about a Name that indicates things like the variable's
        current type and whether that name has been read, set, or
        overwritten.

    Identifier
        A wrapper around variables, used to hold their potential
        non-existence (which is an Issue but not an Error).

    Type
        A symbolic representation of the variable's type.

    Literal
        Sometimes, we need a specialized representation of a literal value
        to be passed around. This is particularly important for accessing
        elements in tuples.

    Name Map
        (Path x Fully Qualified Names) => States
"""
from pedal.core.report import MAIN_REPORT, Report
from pedal.tifa.constants import TOOL_NAME
from pedal.tifa.settings import get_default_tifa_settings
from pedal.tifa.tifa_visitor import Tifa
from pedal.tifa.commands import *
from pedal.types.new_types import reset_builtin_modules


def reset(report=MAIN_REPORT):
    """
    Remove all the information associated with this tool.

    Args:
        report:
    """
    # TODO: Make it so we can reset TIFA through this, safely.
    reset_builtin_modules()
    report[TOOL_NAME] = {
        'analyses': {},
        'latest': None,
        'instance': Tifa(report=report),
        'settings': get_default_tifa_settings(),
        'types': {
            'modules': {}
        }
    }
    return report[TOOL_NAME]


Report.register_tool(TOOL_NAME, reset)


__all__ = ['tifa_analysis', 'Tifa']



from pedal.core.report import MAIN_REPORT


def make_resolver(func, report=None):
    """
    Decorates the given function as a Resolver. This means that when the
    function is executed, the `"pedal.resolver.resolve"` event will be
    triggered.

    Args:
        func (callable): The function to decorate.
        report (Report): The Report to trigger the event on. If None, then use
            the `MAIN_REPORT`.
    """
    if report is None:
        report = MAIN_REPORT

    def resolver_wrapper(*args, **kwargs):
        report.execute_hooks('pedal.resolvers', 'resolve')
        return func(*args, **kwargs)

    return resolver_wrapper



import unittest
import os
import sys
from textwrap import dedent
from contextlib import redirect_stdout
import io

from pedal import verify, run
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

from tests.execution_helper import ExecutionTestCase
from pedal.environments.vpl import VPLEnvironment, VPLFormatter, resolve
from pedal.source import next_section, verify_section, check_section_exists, set_source_file, separate_into_sections
from pedal.core.commands import compliment, clear_report, contextualize_report, log
from pedal.assertions import assert_equal


class TestVPL(ExecutionTestCase):

    def test_vpl_loads_with_syntax_error(self):
        vpl = VPLEnvironment(main_file=here+'datafiles/student_example.py')
        with io.StringIO() as f, redirect_stdout(f):
            vpl.resolve()
            self.assertEqual("""<|--
-Syntax Error
Bad syntax on line 15.

The traceback was:
Line 15 of file datafiles/student_example.py
>a syntax error in this section!
""" + ('>  ^\n' if IS_AT_LEAST_PYTHON_311 else "") +"""
Invalid syntax.
Suggestion: Check line 15, the line before it, and the line after it. Ignore blank lines.
--|>
Grade :=>> 0
""", f.getvalue())

    def test_file_not_found(self):
        vpl = VPLEnvironment(main_file=here + 'datafiles/banana_cream_pudding.py')
        with io.StringIO() as f, redirect_stdout(f):
            vpl.resolve()
            self.assertEqual("""<|--
-Source File Not Found
The given filename datafiles/banana_cream_pudding.py was either not found or could not be opened. Please make sure the file is available.
--|>
Grade :=>> 0
""", f.getvalue())

    def test_simple_system(self):
        clear_report()
        vpl = VPLEnvironment(main_code='1+2')
        with io.StringIO() as f, redirect_stdout(f):
            log("Message Printed")
            assert_equal(1, 2)
            vpl.resolve()
            self.assertEqual("""<|--
-System Notes
Message Printed
-Failed Instructor Test
Student code failed instructor test.
>1 != 2
--|>
Grade :=>> 0
""", f.getvalue())

    def test_simple_system_correct(self):
            clear_report()
            vpl = VPLEnvironment(main_code='1+2')
            with io.StringIO() as f, redirect_stdout(f):
                log("Message Printed")
                vpl.resolve()
                self.assertEqual("""<|--
-System Notes
Message Printed
-Complete
Great work!
--|>
Grade :=>> 1
""", f.getvalue())

    def test_resolve(self):
        clear_report()
        vpl = VPLEnvironment(main_file=here+'datafiles/student_example.py')
        separate_into_sections()
        # Part 0 - usually skipped
        # Part 1
        vpl.next_section()
        # Part 2
        vpl.next_section()
        compliment('Hey, not a bad job!')
        # Part 3
        vpl.next_section()
        # Part 4
        vpl.next_section()
        # Resolve entire thing
        with io.StringIO() as f, redirect_stdout(f):
            vpl.sectional_resolve()
            self.assertEqual("""<|--
-Part 0
No feedback for this section
-Part 1
No feedback for this section
-Part 2
Complete
Great work!
-Part 3
Syntax Error
Bad syntax on line 15.

The traceback was:
Line 15 of file datafiles/student_example.py
>a syntax error in this section!
""" + ('>  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") +"""
Invalid syntax.
Suggestion: Check line 15, the line before it, and the line after it. Ignore blank lines.
-Part 4
No feedback for this section
--|>
Grade :=>> 1
""", f.getvalue())


if __name__ == '__main__':
    unittest.main(buffer=False)



""" Utilities related to file handling """
import os


def normalize_path(first: str, second: str):
    """
    If the `first` path is relative, creates an absolute path relative to the `second` folder.

    Args:
        first: The path to normalize
        second: The path to normalize in regards to.

    Returns:
        str: A new path.
    """
    if os.path.isabs(first):
        return first
    else:
        return os.path.join(os.path.dirname(second), first)


def get_extension(filename: str, default_extension="py") -> str:
    """ Get a file's extension (assumed to be after a period) """
    if not filename:
        return default_extension
    return filename.lower().split(".")[-1]


def find_possible_filenames(filenames, wildcard="*.py", alt_separator=";"):
    """ Yield all possible filenames that we could get from the given single/list of semicolon-separated string."""
    if isinstance(filenames, str):
        if alt_separator in filenames:
            filenames = filenames.split(alt_separator)
        else:
            filenames = [filenames]
    for a_filename in filenames:
        if a_filename.lower().strip().endswith(wildcard):
            a_filename = a_filename[:-len(wildcard)] or "."
            potentials = [f for f in os.listdir(a_filename) if f.endswith(wildcard)]
            if not potentials:
                continue
            a_filename = os.path.join(a_filename, potentials[0])
            yield a_filename
        else:
            yield a_filename




"""
Check utilities.sorting
"""
import sys
import os
import unittest
from pedal.utilities.progsnap import SqlProgSnap2

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

class TestProgsnap(unittest.TestCase):
    def test_progsnap_sort(self):
        progsnap = SqlProgSnap2(here+"datafiles/progsnap2_3.db")
        link_filters = {
            'Subject': {
                'X-IsStaff': "False",
            },
            'Assignment': {
                'X-Name': "Fun%"
            },
        }
        link_selections = {
                'Subject': {
                    'X-Email': 'student_email',
                    'X-Name.First': 'student_first',
                    'X-Name.Last': 'student_last',
                },
                'Assignment': {
                    'X-Name': 'assignment_name',
                    'X-URL': 'assignment_url',
                    'X-Code.OnRun': 'on_run'
                }
            }
        fun_student_edits = progsnap.get_events(
            event_filter={'EventType': 'File.Edit'},
            link_filters=link_filters,
            link_selections=link_selections
        )
        self.assertEqual(212, len(fun_student_edits))
        self.assertIsInstance(fun_student_edits[0], dict)
        self.assertEqual("592", fun_student_edits[0]['event_id'])


if __name__ == '__main__':
    unittest.main(buffer=False)



from pedal.types.new_types import ModuleType, FunctionType, NoneType, register_builtin_module


# MatPlotLib Support
def load_matplot_module():
    _PYPLOT_MODULE = ModuleType('pyplot', fields={
        'plot': FunctionType(name='plot', returns=NoneType),
        'hist': FunctionType(name='hist', returns=NoneType),
        'scatter': FunctionType(name='scatter', returns=NoneType),
        'show': FunctionType(name='show', returns=NoneType),
        'xlabel': FunctionType(name='xlabel', returns=NoneType),
        'ylabel': FunctionType(name='ylabel', returns=NoneType),
        'title': FunctionType(name='title', returns=NoneType),
    })
    return ModuleType('matplotlib', {}, {'pyplot': _PYPLOT_MODULE})


register_builtin_module('matplotlib', load_matplot_module)



"""
Utilities for handling English Language rules.

"""


def add_indefinite_article(phrase):
    """
    Given a word, choose the appropriate indefinite article (e.g., "a" or "an")
    and add it the front of the word. Matches the original word's
    capitalization.


    Args:
        phrase (str): The phrase to get the appropriate definite article
            for.

    Returns:
        str: Either "an" or "a".
    """
    # Note: Must cast to string because it could be a SandboxResult
    if str(phrase[0]) in "aeiou":
        return "an "+phrase
    elif str(phrase[0]) in "AEIOU":
        return "An "+phrase
    elif phrase[0].lower() == phrase[0]:
        return "a "+phrase
    else:
        return "A "+phrase


def escape_curly_braces(result):
    """ Replaces all occurrences of curly braces with double curly braces. """
    return result.replace("{", "{{").replace("}", "}}")


def safe_repr(obj, short=False, max_length=80) -> str:
    """
    Create a string representation of this object using :py:func:`repr`. If the
    object doesn't have a repr defined, then falls back onto the default
    :py:func:`object.__repr__`. Finally, if ``short`` is true, then the string
    will be truncated to the max_length value.

    Args:
        obj (Any): Any object to repr.
        short (bool): Whether or not to truncate to the max_length.
        max_length (int): How many maximum characters to use, if short is True.
    """
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)
    if short and len(result) >= max_length:
        result = result[:max_length] + ' [truncated]...'
    result = result
    return result


def chomp(text):
    """ Removes the trailing newline character, if there is one. """
    if not text:
        return text
    if text[-2:] == '\n\r':
        return text[:-2]
    if text[-1] == '\n':
        return text[:-1]
    return text


def join_list_with_and(items):
    """ Properly format the string with commas (oxford-style) and the word `and`."""
    items = list(items)
    if len(items) < 1:
        return ""
    elif len(items) == 1:
        return items[0]
    elif len(items) == 2:
        return f"{items[0]} and {items[1]}"
    else:
        pre = ", ".join(items[:-1])
        return f"{pre}, and {items[-1]}"

def inject_line(code, line_number, new_line):
    lines = code.split("\n")
    lines.insert(line_number, new_line)
    return "\n".join(lines)


def render_table(rows, headers=None, separator=" | ", max_width=120):
    """
    Simple function to render table with fixed width.

    TODO: Try to resize table to make it fit comfortably within 120 characters,
        e.g., by removing separators, or by wrapping text.

    Based on: https://stackoverflow.com/a/52247284/1718155
    """
    table = [headers] + rows if headers else rows
    longest_cols = [
        (max([len(str(row[i])) for row in table]))
        for i in range(len(table[0]))
    ]
    total_width = sum(longest_cols) + (len(longest_cols)-1) * len(separator)
    if max_width is not None:
        if total_width > max_width:
            # Maybe we can trim the separator?
            if separator:
                shorter_separator = separator.strip()
                # Oops, we already tried stripping. What about removing it?
                if shorter_separator == separator:
                    shorter_separator = ""
                return render_table(rows, headers, shorter_separator, max_width)
    row_format = separator.join(["{:>" + str(longest_col) + "}"
                                for longest_col in longest_cols])
    result = [row_format.format(*row) for row in table]
    if headers:
        result.insert(1, "="*total_width)
    return "\n".join(result)
